=== modified file 'bzr'
--- bzr	2007-10-16 16:02:01 +0000
+++ bzr	2007-10-22 01:23:51 +0000
@@ -94,6 +94,7 @@
 import bzrlib.commands
 import bzrlib.trace
 
+
 if __name__ == '__main__':
     bzrlib.trace.enable_default_logging()
     exit_val = bzrlib.commands.main(sys.argv)

=== modified file 'bzrlib/bzrdir.py'
--- bzrlib/bzrdir.py	2007-09-21 07:29:37 +0000
+++ bzrlib/bzrdir.py	2007-10-23 06:57:05 +0000
@@ -2478,4 +2478,24 @@
     tree_format='bzrlib.workingtree.WorkingTreeFormat4',
     hidden=True,
     )
+format_registry.register_metadir('experimental',
+    'bzrlib.repofmt.pack_repo.RepositoryFormatKnitPack1',
+    help='New in XXX: Experimental format with data compatible with dirstate '
+        'format repositories. Cannot be read except with bzr.dev. '
+        'WARNING: This format is unstable and data in it will not be upgradable'
+        ' to release formats of bzr.',
+    branch_format='bzrlib.branch.BzrBranchFormat6',
+    tree_format='bzrlib.workingtree.WorkingTreeFormat4',
+    hidden=True,
+    )
+format_registry.register_metadir('experimental-subtree',
+    'bzrlib.repofmt.pack_repo.RepositoryFormatKnitPack3',
+    help='New in XXX: Experimental format with data compatible with '
+        'dirstate-with-subtree format repositories. Cannot be read except with'
+        ' bzr.dev. WARNING: This format is unstable and data in it will not be'
+        ' upgradable to release formats of bzr.',
+    branch_format='bzrlib.branch.BzrBranchFormat6',
+    tree_format='bzrlib.workingtree.WorkingTreeFormat4',
+    hidden=True,
+    )
 format_registry.set_default('dirstate-tags')

=== modified file 'bzrlib/delta.py'
--- bzrlib/delta.py	2007-10-16 16:02:01 +0000
+++ bzrlib/delta.py	2007-10-24 02:33:14 +0000
@@ -333,7 +333,7 @@
                 # on a rename, we show old and new
                 old_path, path = paths
             else:
-                # if its not renamed, we're showing both for kind changes
+                # if it's not renamed, we're showing both for kind changes
                 # so only show the new path
                 old_path, path = paths[1], paths[1]
             # if the file is not missing in the source, we show its kind

=== modified file 'bzrlib/graph.py'
--- bzrlib/graph.py	2007-10-22 19:49:31 +0000
+++ bzrlib/graph.py	2007-10-24 06:48:13 +0000
@@ -369,6 +369,30 @@
             return set(heads)
 
 
+class HeadsCache(object):
+    """A cache of results for graph heads calls."""
+
+    def __init__(self, graph):
+        self.graph = graph
+        self._heads = {}
+
+    def heads(self, keys):
+        """Return the heads of keys.
+
+        :see also: Graph.heads.
+        :param keys: The keys to calculate heads for.
+        :return: A set containing the heads, which may be mutated without
+            affecting future lookups.
+        """
+        keys = frozenset(keys)
+        try:
+            return set(self._heads[keys])
+        except KeyError:
+            heads = self.graph.heads(keys)
+            self._heads[keys] = heads
+            return set(heads)
+
+
 class _BreadthFirstSearcher(object):
     """Parallel search breadth-first the ancestry of revisions.
 

=== modified file 'bzrlib/index.py'
--- bzrlib/index.py	2007-10-18 01:12:22 +0000
+++ bzrlib/index.py	2007-10-18 01:17:01 +0000
@@ -267,6 +267,17 @@
         self._nodes_by_key = None
         self._size = size
 
+    def __eq__(self, other):
+        """Equal when self and other were created with the same parameters."""
+        return (
+            type(self) == type(other) and
+            self._transport == other._transport and
+            self._name == other._name and
+            self._size == other._size)
+
+    def __ne__(self, other):
+        return not self.__eq__(other)
+
     def _buffer_all(self):
         """Buffer all the index data.
 

=== modified file 'bzrlib/knit.py'
--- bzrlib/knit.py	2007-10-22 19:53:26 +0000
+++ bzrlib/knit.py	2007-10-24 06:48:13 +0000
@@ -1984,7 +1984,8 @@
 
     def set_writer(self, writer, index, (transport, packname)):
         """Set a writer to use for adding data."""
-        self.indices[index] = (transport, packname)
+        if index is not None:
+            self.indices[index] = (transport, packname)
         self.container_writer = writer
         self.write_index = index
 

=== modified file 'bzrlib/reconcile.py'
--- bzrlib/reconcile.py	2007-10-15 09:10:30 +0000
+++ bzrlib/reconcile.py	2007-10-24 05:33:57 +0000
@@ -17,19 +17,22 @@
 """Reconcilers are able to fix some potential data errors in a branch."""
 
 
-__all__ = ['reconcile', 'Reconciler', 'RepoReconciler', 'KnitReconciler']
+__all__ = [
+    'KnitReconciler',
+    'PackReconciler',
+    'reconcile',
+    'Reconciler',
+    'RepoReconciler',
+    ]
 
 
 from bzrlib import (
     errors,
-    graph,
     ui,
     repository,
     )
-from bzrlib import errors
-from bzrlib import ui
 from bzrlib.trace import mutter, note
-from bzrlib.tsort import TopoSorter, topo_sort
+from bzrlib.tsort import TopoSorter
 
 
 def reconcile(dir, other=None):
@@ -397,9 +400,30 @@
             else:
                 parents = vf.get_parents(version)
             new_parents[version] = parents
-        for version in topo_sort(new_parents.items()):
+        for version in TopoSorter(new_parents.items()).iter_topo_order():
             new_vf.add_lines(version, new_parents[version],
                              vf.get_lines(version))
         self.repo.weave_store.copy(new_vf, file_id, self.transaction)
         self.repo.weave_store.delete('temp:%s' % file_id, self.transaction)
 
+
+class PackReconciler(RepoReconciler):
+    """Reconciler that reconciles a pack based repository.
+
+    Garbage inventories do not affect ancestry queries, and removal is
+    considerably more expensive as there is no separate versioned file for
+    them, so they are not cleaned. In short it is currently a no-op.
+
+    In future this may be a good place to hook in annotation cache checking,
+    index recreation etc.
+    """
+
+    # XXX: The index corruption that _fix_text_parents performs is needed for
+    # packs, but not yet implemented. The basic approach is to:
+    #  - lock the names list
+    #  - perform a customised pack() that regenerates data as needed
+    #  - unlock the names list
+    # https://bugs.edge.launchpad.net/bzr/+bug/154173
+
+    def _reconcile_steps(self):
+        """Perform the steps to reconcile this repository."""

=== modified file 'bzrlib/repofmt/knitrepo.py'
--- bzrlib/repofmt/knitrepo.py	2007-10-12 08:18:54 +0000
+++ bzrlib/repofmt/knitrepo.py	2007-10-17 09:39:41 +0000
@@ -276,14 +276,13 @@
         :returns: an iterator yielding tuples of (revison-id, parents-in-index,
             parents-in-revision).
         """
+        assert self.is_locked()
         vf = self._get_revision_vf()
-        index_versions = vf.versions()
-        for index_version in index_versions:
-            parents_according_to_index = vf._index.get_parents_with_ghosts(
-                index_version)
-            revision = self._revision_store.get_revision(index_version,
-                self.get_transaction())
-            parents_according_to_revision = revision.parent_ids
+        for index_version in vf.versions():
+            parents_according_to_index = tuple(vf.get_parents_with_ghosts(
+                index_version))
+            revision = self.get_revision(index_version)
+            parents_according_to_revision = tuple(revision.parent_ids)
             if parents_according_to_index != parents_according_to_revision:
                 yield (index_version, parents_according_to_index,
                     parents_according_to_revision)

=== added file 'bzrlib/repofmt/pack_repo.py'
--- bzrlib/repofmt/pack_repo.py	1970-01-01 00:00:00 +0000
+++ bzrlib/repofmt/pack_repo.py	2007-10-24 05:17:39 +0000
@@ -0,0 +1,1727 @@
+# Copyright (C) 2005, 2006, 2007 Canonical Ltd
+#
+# This program is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation; either version 2 of the License, or
+# (at your option) any later version.
+#
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU General Public License for more details.
+#
+# You should have received a copy of the GNU General Public License
+# along with this program; if not, write to the Free Software
+# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+
+from bzrlib.lazy_import import lazy_import
+lazy_import(globals(), """
+from itertools import izip
+import math
+import md5
+import time
+
+from bzrlib import (
+        debug,
+        pack,
+        ui,
+        )
+from bzrlib.index import (
+    GraphIndex,
+    GraphIndexBuilder,
+    InMemoryGraphIndex,
+    CombinedGraphIndex,
+    GraphIndexPrefixAdapter,
+    )
+from bzrlib.knit import KnitGraphIndex, _PackAccess, _KnitData
+from bzrlib.osutils import rand_chars
+from bzrlib.pack import ContainerWriter
+from bzrlib.store import revision
+""")
+from bzrlib import (
+    bzrdir,
+    deprecated_graph,
+    errors,
+    knit,
+    lockable_files,
+    lockdir,
+    osutils,
+    transactions,
+    xml5,
+    xml7,
+    )
+
+from bzrlib.decorators import needs_read_lock, needs_write_lock
+from bzrlib.repofmt.knitrepo import KnitRepository
+from bzrlib.repository import (
+    CommitBuilder,
+    MetaDirRepository,
+    MetaDirRepositoryFormat,
+    RootCommitBuilder,
+    )
+import bzrlib.revision as _mod_revision
+from bzrlib.store.revision.knit import KnitRevisionStore
+from bzrlib.store.versioned import VersionedFileStore
+from bzrlib.trace import mutter, note, warning
+
+
+class PackCommitBuilder(CommitBuilder):
+    """A subclass of CommitBuilder to add texts with pack semantics.
+    
+    Specifically this uses one knit object rather than one knit object per
+    added text, reducing memory and object pressure.
+    """
+
+    def _add_text_to_weave(self, file_id, new_lines, parents, nostore_sha):
+        return self.repository._pack_collection._add_text_to_weave(file_id,
+            self._new_revision_id, new_lines, parents, nostore_sha,
+            self.random_revid)
+
+
+class PackRootCommitBuilder(RootCommitBuilder):
+    """A subclass of RootCommitBuilder to add texts with pack semantics.
+    
+    Specifically this uses one knit object rather than one knit object per
+    added text, reducing memory and object pressure.
+    """
+
+    def _add_text_to_weave(self, file_id, new_lines, parents, nostore_sha):
+        return self.repository._pack_collection._add_text_to_weave(file_id,
+            self._new_revision_id, new_lines, parents, nostore_sha,
+            self.random_revid)
+
+
+class Pack(object):
+    """An in memory proxy for a pack and its indices.
+
+    This is a base class that is not directly used, instead the classes
+    ExistingPack and NewPack are used.
+    """
+
+    def __init__(self, revision_index, inventory_index, text_index,
+        signature_index):
+        """Create a pack instance.
+
+        :param revision_index: A GraphIndex for determining what revisions are
+            present in the Pack and accessing the locations of their texts.
+        :param inventory_index: A GraphIndex for determining what inventories are
+            present in the Pack and accessing the locations of their
+            texts/deltas.
+        :param text_index: A GraphIndex for determining what file texts
+            are present in the pack and accessing the locations of their
+            texts/deltas (via (fileid, revisionid) tuples).
+        :param revision_index: A GraphIndex for determining what signatures are
+            present in the Pack and accessing the locations of their texts.
+        """
+        self.revision_index = revision_index
+        self.inventory_index = inventory_index
+        self.text_index = text_index
+        self.signature_index = signature_index
+
+    def access_tuple(self):
+        """Return a tuple (transport, name) for the pack content."""
+        return self.pack_transport, self.file_name()
+
+    def file_name(self):
+        """Get the file name for the pack on disk."""
+        return self.name + '.pack'
+
+    def get_revision_count(self):
+        return self.revision_index.key_count()
+
+    def inventory_index_name(self, name):
+        """The inv index is the name + .iix."""
+        return self.index_name('inventory', name)
+
+    def revision_index_name(self, name):
+        """The revision index is the name + .rix."""
+        return self.index_name('revision', name)
+
+    def signature_index_name(self, name):
+        """The signature index is the name + .six."""
+        return self.index_name('signature', name)
+
+    def text_index_name(self, name):
+        """The text index is the name + .tix."""
+        return self.index_name('text', name)
+
+
+class ExistingPack(Pack):
+    """An in memory proxy for an existing .pack and its disk indices."""
+
+    def __init__(self, pack_transport, name, revision_index, inventory_index,
+        text_index, signature_index):
+        """Create an ExistingPack object.
+
+        :param pack_transport: The transport where the pack file resides.
+        :param name: The name of the pack on disk in the pack_transport.
+        """
+        Pack.__init__(self, revision_index, inventory_index, text_index,
+            signature_index)
+        self.name = name
+        self.pack_transport = pack_transport
+        assert None not in (revision_index, inventory_index, text_index,
+            signature_index, name, pack_transport)
+
+    def __eq__(self, other):
+        return self.__dict__ == other.__dict__
+
+    def __ne__(self, other):
+        return not self.__eq__(other)
+
+    def __repr__(self):
+        return "<bzrlib.repofmt.pack_repo.Pack object at 0x%x, %s, %s" % (
+            id(self), self.transport, self.name)
+
+
+class NewPack(Pack):
+    """An in memory proxy for a pack which is being created."""
+
+    # A map of index 'type' to the file extension and position in the
+    # index_sizes array.
+    index_definitions = {
+        'revision': ('.rix', 0),
+        'inventory': ('.iix', 1),
+        'text': ('.tix', 2),
+        'signature': ('.six', 3),
+        }
+
+    def __init__(self, upload_transport, index_transport, pack_transport,
+        upload_suffix=''):
+        """Create a NewPack instance.
+
+        :param upload_transport: A writable transport for the pack to be
+            incrementally uploaded to.
+        :param index_transport: A writable transport for the pack's indices to
+            be written to when the pack is finished.
+        :param pack_transport: A writable transport for the pack to be renamed
+            to when the upload is complete. This *must* be the same as
+            upload_transport.clone('../packs').
+        :param upload_suffix: An optional suffix to be given to any temporary
+            files created during the pack creation. e.g '.autopack'
+        """
+        # The relative locations of the packs are constrained, but all are
+        # passed in because the caller has them, so as to avoid object churn.
+        Pack.__init__(self,
+            # Revisions: parents list, no text compression.
+            InMemoryGraphIndex(reference_lists=1),
+            # Inventory: We want to map compression only, but currently the
+            # knit code hasn't been updated enough to understand that, so we
+            # have a regular 2-list index giving parents and compression
+            # source.
+            InMemoryGraphIndex(reference_lists=2),
+            # Texts: compression and per file graph, for all fileids - so two
+            # reference lists and two elements in the key tuple.
+            InMemoryGraphIndex(reference_lists=2, key_elements=2),
+            # Signatures: Just blobs to store, no compression, no parents
+            # listing.
+            InMemoryGraphIndex(reference_lists=0),
+            )
+        # where should the new pack be opened
+        self.upload_transport = upload_transport
+        # where are indices written out to
+        self.index_transport = index_transport
+        # where is the pack renamed to when it is finished?
+        self.pack_transport = pack_transport
+        # tracks the content written to the .pack file.
+        self._hash = md5.new()
+        # a four-tuple with the length in bytes of the indices, once the pack
+        # is finalised. (rev, inv, text, sigs)
+        self.index_sizes = None
+        # How much data to cache when writing packs. Note that this is not
+        # synchronised with reads, because it's not in the transport layer, so
+        # is not safe unless the client knows it won't be reading from the pack
+        # under creation.
+        self._cache_limit = 0
+        # the temporary pack file name.
+        self.random_name = rand_chars(20) + upload_suffix
+        # when was this pack started ?
+        self.start_time = time.time()
+        # open an output stream for the data added to the pack.
+        self.write_stream = self.upload_transport.open_write_stream(
+            self.random_name)
+        if 'pack' in debug.debug_flags:
+            mutter('%s: create_pack: pack stream open: %s%s t+%6.3fs',
+                time.ctime(), self.upload_transport.base, self.random_name,
+                time.time() - self.start_time)
+        # A list of byte sequences to be written to the new pack, and the 
+        # aggregate size of them.  Stored as a list rather than separate 
+        # variables so that the _write_data closure below can update them.
+        self._buffer = [[], 0]
+        # create a callable for adding data 
+        #
+        # robertc says- this is a closure rather than a method on the object
+        # so that the variables are locals, and faster than accessing object
+        # members.
+        def _write_data(bytes, flush=False, _buffer=self._buffer,
+            _write=self.write_stream.write, _update=self._hash.update):
+            _buffer[0].append(bytes)
+            _buffer[1] += len(bytes)
+            # buffer cap
+            if _buffer[1] > self._cache_limit or flush:
+                bytes = ''.join(_buffer[0])
+                _write(bytes)
+                _update(bytes)
+                _buffer[:] = [[], 0]
+        # expose this on self, for the occasion when clients want to add data.
+        self._write_data = _write_data
+        # a pack writer object to serialise pack records.
+        self._writer = pack.ContainerWriter(self._write_data)
+        self._writer.begin()
+        # what state is the pack in? (open, finished, aborted)
+        self._state = 'open'
+
+    def abort(self):
+        """Cancel creating this pack."""
+        self._state = 'aborted'
+        # Remove the temporary pack file.
+        self.upload_transport.delete(self.random_name)
+        # The indices have no state on disk.
+
+    def access_tuple(self):
+        """Return a tuple (transport, name) for the pack content."""
+        assert self._state in ('open', 'finished')
+        if self._state == 'finished':
+            return Pack.access_tuple(self)
+        else:
+            return self.upload_transport, self.random_name
+
+    def data_inserted(self):
+        """True if data has been added to this pack."""
+        return bool(self.get_revision_count() or
+            self.inventory_index.key_count() or
+            self.text_index.key_count() or
+            self.signature_index.key_count())
+
+    def finish(self):
+        """Finish the new pack.
+
+        This:
+         - finalises the content
+         - assigns a name (the md5 of the content, currently)
+         - writes out the associated indices
+         - renames the pack into place.
+         - stores the index size tuple for the pack in the index_sizes
+           attribute.
+        """
+        self._writer.end()
+        if self._buffer[1]:
+            self._write_data('', flush=True)
+        self.name = self._hash.hexdigest()
+        # write indices
+        # XXX: It'd be better to write them all to temporary names, then
+        # rename them all into place, so that the window when only some are
+        # visible is smaller.  On the other hand none will be seen until
+        # they're in the names list.
+        self.index_sizes = [None, None, None, None]
+        self._write_index('revision', self.revision_index, 'revision')
+        self._write_index('inventory', self.inventory_index, 'inventory')
+        self._write_index('text', self.text_index, 'file texts')
+        self._write_index('signature', self.signature_index,
+            'revision signatures')
+        self.write_stream.close()
+        # Note that this will clobber an existing pack with the same name,
+        # without checking for hash collisions. While this is undesirable this
+        # is something that can be rectified in a subsequent release. One way
+        # to rectify it may be to leave the pack at the original name, writing
+        # its pack-names entry as something like 'HASH: index-sizes
+        # temporary-name'. Allocate that and check for collisions, if it is
+        # collision free then rename it into place. If clients know this scheme
+        # they can handle missing-file errors by:
+        #  - try for HASH.pack
+        #  - try for temporary-name
+        #  - refresh the pack-list to see if the pack is now absent
+        self.upload_transport.rename(self.random_name,
+                '../packs/' + self.name + '.pack')
+        self._state = 'finished'
+        if 'pack' in debug.debug_flags:
+            # XXX: size might be interesting?
+            mutter('%s: create_pack: pack renamed into place: %s%s->%s%s t+%6.3fs',
+                time.ctime(), self.upload_transport.base, self.random_name,
+                self.pack_transport, self.name,
+                time.time() - self.start_time)
+
+    def index_name(self, index_type, name):
+        """Get the disk name of an index type for pack name 'name'."""
+        return name + NewPack.index_definitions[index_type][0]
+
+    def index_offset(self, index_type):
+        """Get the position in a index_size array for a given index type."""
+        return NewPack.index_definitions[index_type][1]
+
+    def _replace_index_with_readonly(self, index_type):
+        setattr(self, index_type + '_index',
+            GraphIndex(self.index_transport,
+                self.index_name(index_type, self.name),
+                self.index_sizes[self.index_offset(index_type)]))
+
+    def set_write_cache_size(self, size):
+        self._cache_limit = size
+
+    def _write_index(self, index_type, index, label):
+        """Write out an index.
+
+        :param index_type: The type of index to write - e.g. 'revision'.
+        :param index: The index object to serialise.
+        :param label: What label to give the index e.g. 'revision'.
+        """
+        index_name = self.index_name(index_type, self.name)
+        self.index_sizes[self.index_offset(index_type)] = \
+            self.index_transport.put_file(index_name, index.finish())
+        if 'pack' in debug.debug_flags:
+            # XXX: size might be interesting?
+            mutter('%s: create_pack: wrote %s index: %s%s t+%6.3fs',
+                time.ctime(), label, self.upload_transport.base,
+                self.random_name, time.time() - self.start_time)
+        # Replace the writable index on this object with a readonly, 
+        # presently unloaded index. We should alter
+        # the index layer to make its finish() error if add_node is
+        # subsequently used. RBC
+        self._replace_index_with_readonly(index_type)
+
+
+class AggregateIndex(object):
+    """An aggregated index for the RepositoryPackCollection.
+
+    AggregateIndex is reponsible for managing the PackAccess object,
+    Index-To-Pack mapping, and all indices list for a specific type of index
+    such as 'revision index'.
+
+    A CombinedIndex provides an index on a single key space built up
+    from several on-disk indices.  The AggregateIndex builds on this 
+    to provide a knit access layer, and allows having up to one writable
+    index within the collection.
+    """
+    # XXX: Probably 'can be written to' could/should be separated from 'acts
+    # like a knit index' -- mbp 20071024
+
+    def __init__(self):
+        """Create an AggregateIndex."""
+        self.index_to_pack = {}
+        self.combined_index = CombinedGraphIndex([])
+        self.knit_access = _PackAccess(self.index_to_pack)
+
+    def replace_indices(self, index_to_pack, indices):
+        """Replace the current mappings with fresh ones.
+
+        This should probably not be used eventually, rather incremental add and
+        removal of indices. It has been added during refactoring of existing
+        code.
+
+        :param index_to_pack: A mapping from index objects to
+            (transport, name) tuples for the pack file data.
+        :param indices: A list of indices.
+        """
+        # refresh the revision pack map dict without replacing the instance.
+        self.index_to_pack.clear()
+        self.index_to_pack.update(index_to_pack)
+        # XXX: API break - clearly a 'replace' method would be good?
+        self.combined_index._indices[:] = indices
+        # the current add nodes callback for the current writable index if
+        # there is one.
+        self.add_callback = None
+
+    def add_index(self, index, pack):
+        """Add index to the aggregate, which is an index for Pack pack.
+
+        Future searches on the aggregate index will seach this new index
+        before all previously inserted indices.
+        
+        :param index: An Index for the pack.
+        :param pack: A Pack instance.
+        """
+        # expose it to the index map
+        self.index_to_pack[index] = pack.access_tuple()
+        # put it at the front of the linear index list
+        self.combined_index.insert_index(0, index)
+
+    def add_writable_index(self, index, pack):
+        """Add an index which is able to have data added to it.
+
+        There can be at most one writable index at any time.  Any
+        modifications made to the knit are put into this index.
+        
+        :param index: An index from the pack parameter.
+        :param pack: A Pack instance.
+        """
+        assert self.add_callback is None, \
+            "%s already has a writable index through %s" % \
+            (self, self.add_callback)
+        # allow writing: queue writes to a new index
+        self.add_index(index, pack)
+        # Updates the index to packs mapping as a side effect,
+        self.knit_access.set_writer(pack._writer, index, pack.access_tuple())
+        self.add_callback = index.add_nodes
+
+    def clear(self):
+        """Reset all the aggregate data to nothing."""
+        self.knit_access.set_writer(None, None, (None, None))
+        self.index_to_pack.clear()
+        del self.combined_index._indices[:]
+        self.add_callback = None
+
+    def remove_index(self, index, pack):
+        """Remove index from the indices used to answer queries.
+        
+        :param index: An index from the pack parameter.
+        :param pack: A Pack instance.
+        """
+        del self.index_to_pack[index]
+        self.combined_index._indices.remove(index)
+        if (self.add_callback is not None and
+            getattr(index, 'add_nodes', None) == self.add_callback):
+            self.add_callback = None
+            self.knit_access.set_writer(None, None, (None, None))
+
+
+class RepositoryPackCollection(object):
+    """Management of packs within a repository."""
+
+    def __init__(self, repo, transport, index_transport, upload_transport,
+                 pack_transport):
+        """Create a new RepositoryPackCollection.
+
+        :param transport: Addresses the repository base directory 
+            (typically .bzr/repository/).
+        :param index_transport: Addresses the directory containing indices.
+        :param upload_transport: Addresses the directory into which packs are written
+            while they're being created.
+        :param pack_transport: Addresses the directory of existing complete packs.
+        """
+        self.repo = repo
+        self.transport = transport
+        self._index_transport = index_transport
+        self._upload_transport = upload_transport
+        self._pack_transport = pack_transport
+        self._suffix_offsets = {'.rix': 0, '.iix': 1, '.tix': 2, '.six': 3}
+        self.packs = []
+        # name:Pack mapping
+        self._packs_by_name = {}
+        # the previous pack-names content
+        self._packs_at_load = None
+        # when a pack is being created by this object, the state of that pack.
+        self._new_pack = None
+        # aggregated revision index data
+        self.revision_index = AggregateIndex()
+        self.inventory_index = AggregateIndex()
+        self.text_index = AggregateIndex()
+        self.signature_index = AggregateIndex()
+
+    def add_pack_to_memory(self, pack):
+        """Make a Pack object available to the repository to satisfy queries.
+        
+        :param pack: A Pack object.
+        """
+        assert pack.name not in self._packs_by_name
+        self.packs.append(pack)
+        self._packs_by_name[pack.name] = pack
+        self.revision_index.add_index(pack.revision_index, pack)
+        self.inventory_index.add_index(pack.inventory_index, pack)
+        self.text_index.add_index(pack.text_index, pack)
+        self.signature_index.add_index(pack.signature_index, pack)
+        
+    def _add_text_to_weave(self, file_id, revision_id, new_lines, parents,
+        nostore_sha, random_revid):
+        file_id_index = GraphIndexPrefixAdapter(
+            self.text_index.combined_index,
+            (file_id, ), 1,
+            add_nodes_callback=self.text_index.add_callback)
+        self.repo._text_knit._index._graph_index = file_id_index
+        self.repo._text_knit._index._add_callback = file_id_index.add_nodes
+        return self.repo._text_knit.add_lines_with_ghosts(
+            revision_id, parents, new_lines, nostore_sha=nostore_sha,
+            random_id=random_revid, check_content=False)[0:2]
+
+    def all_packs(self):
+        """Return a list of all the Pack objects this repository has.
+
+        Note that an in-progress pack being created is not returned.
+
+        :return: A list of Pack objects for all the packs in the repository.
+        """
+        result = []
+        for name in self.names():
+            result.append(self.get_pack_by_name(name))
+        return result
+
+    def autopack(self):
+        """Pack the pack collection incrementally.
+        
+        This will not attempt global reorganisation or recompression,
+        rather it will just ensure that the total number of packs does
+        not grow without bound. It uses the _max_pack_count method to
+        determine if autopacking is needed, and the pack_distribution
+        method to determine the number of revisions in each pack.
+
+        If autopacking takes place then the packs name collection will have
+        been flushed to disk - packing requires updating the name collection
+        in synchronisation with certain steps. Otherwise the names collection
+        is not flushed.
+
+        :return: True if packing took place.
+        """
+        # XXX: Should not be needed when the management of indices is sane.
+        total_revisions = self.revision_index.combined_index.key_count()
+        total_packs = len(self._names)
+        if self._max_pack_count(total_revisions) >= total_packs:
+            return False
+        # XXX: the following may want to be a class, to pack with a given
+        # policy.
+        mutter('Auto-packing repository %s, which has %d pack files, '
+            'containing %d revisions into %d packs.', self, total_packs,
+            total_revisions, self._max_pack_count(total_revisions))
+        # determine which packs need changing
+        pack_distribution = self.pack_distribution(total_revisions)
+        existing_packs = []
+        for pack in self.all_packs():
+            revision_count = pack.get_revision_count()
+            if revision_count == 0:
+                # revision less packs are not generated by normal operation,
+                # only by operations like sign-my-commits, and thus will not
+                # tend to grow rapdily or without bound like commit containing
+                # packs do - leave them alone as packing them really should
+                # group their data with the relevant commit, and that may
+                # involve rewriting ancient history - which autopack tries to
+                # avoid. Alternatively we could not group the data but treat
+                # each of these as having a single revision, and thus add 
+                # one revision for each to the total revision count, to get
+                # a matching distribution.
+                continue
+            existing_packs.append((revision_count, pack))
+        pack_operations = self.plan_autopack_combinations(
+            existing_packs, pack_distribution)
+        self._execute_pack_operations(pack_operations)
+        return True
+
+    def create_pack_from_packs(self, packs, suffix, revision_ids=None):
+        """Create a new pack by reading data from other packs.
+
+        This does little more than a bulk copy of data. One key difference
+        is that data with the same item key across multiple packs is elided
+        from the output. The new pack is written into the current pack store
+        along with its indices, and the name added to the pack names. The 
+        source packs are not altered and are not required to be in the current
+        pack collection.
+
+        :param packs: An iterable of Packs to combine.
+        :param revision_ids: Either None, to copy all data, or a list
+            of revision_ids to limit the copied data to the data they
+            introduced.
+        :return: A Pack object, or None if nothing was copied.
+        """
+        # open a pack - using the same name as the last temporary file
+        # - which has already been flushed, so its safe.
+        # XXX: - duplicate code warning with start_write_group; fix before
+        #      considering 'done'.
+        if self._new_pack is not None:
+            raise errors.BzrError('call to create_pack_from_packs while '
+                'another pack is being written.')
+        if revision_ids is not None and len(revision_ids) == 0:
+            # silly fetch request.
+            return None
+        new_pack = NewPack(self._upload_transport, self._index_transport,
+            self._pack_transport, upload_suffix=suffix)
+        # buffer data - we won't be reading-back during the pack creation and
+        # this makes a significant difference on sftp pushes.
+        new_pack.set_write_cache_size(1024*1024)
+        if 'pack' in debug.debug_flags:
+            plain_pack_list = ['%s%s' % (a_pack.pack_transport.base, a_pack.name)
+                for a_pack in packs]
+            if revision_ids is not None:
+                rev_count = len(revision_ids)
+            else:
+                rev_count = 'all'
+            mutter('%s: create_pack: creating pack from source packs: '
+                '%s%s %s revisions wanted %s t=0',
+                time.ctime(), self._upload_transport.base, new_pack.random_name,
+                plain_pack_list, rev_count)
+        # select revisions
+        if revision_ids:
+            revision_keys = [(revision_id,) for revision_id in revision_ids]
+        else:
+            revision_keys = None
+
+        # select revision keys
+        revision_index_map = self._packs_list_to_pack_map_and_index_list(
+            packs, 'revision_index')[0]
+        revision_nodes = self._index_contents(revision_index_map, revision_keys)
+        # copy revision keys and adjust values
+        list(self._copy_nodes_graph(revision_nodes, revision_index_map,
+            new_pack._writer, new_pack.revision_index))
+        if 'pack' in debug.debug_flags:
+            mutter('%s: create_pack: revisions copied: %s%s %d items t+%6.3fs',
+                time.ctime(), self._upload_transport.base, new_pack.random_name,
+                new_pack.revision_index.key_count(),
+                time.time() - new_pack.start_time)
+        # select inventory keys
+        inv_keys = revision_keys # currently the same keyspace, and note that
+        # querying for keys here could introduce a bug where an inventory item
+        # is missed, so do not change it to query separately without cross
+        # checking like the text key check below.
+        inventory_index_map = self._packs_list_to_pack_map_and_index_list(
+            packs, 'inventory_index')[0]
+        inv_nodes = self._index_contents(inventory_index_map, inv_keys)
+        # copy inventory keys and adjust values
+        # XXX: Should be a helper function to allow different inv representation
+        # at this point.
+        inv_lines = self._copy_nodes_graph(inv_nodes, inventory_index_map,
+            new_pack._writer, new_pack.inventory_index, output_lines=True)
+        if revision_ids:
+            fileid_revisions = self.repo._find_file_ids_from_xml_inventory_lines(
+                inv_lines, revision_ids)
+            text_filter = []
+            for fileid, file_revids in fileid_revisions.iteritems():
+                text_filter.extend(
+                    [(fileid, file_revid) for file_revid in file_revids])
+        else:
+            # eat the iterator to cause it to execute.
+            list(inv_lines)
+            text_filter = None
+        if 'pack' in debug.debug_flags:
+            mutter('%s: create_pack: inventories copied: %s%s %d items t+%6.3fs',
+                time.ctime(), self._upload_transport.base, new_pack.random_name,
+                new_pack.inventory_index.key_count(),
+                time.time() - new_pack.start_time)
+        # select text keys
+        text_index_map = self._packs_list_to_pack_map_and_index_list(
+            packs, 'text_index')[0]
+        text_nodes = self._index_contents(text_index_map, text_filter)
+        if text_filter is not None:
+            # We could return the keys copied as part of the return value from
+            # _copy_nodes_graph but this doesn't work all that well with the
+            # need to get line output too, so we check separately, and as we're
+            # going to buffer everything anyway, we check beforehand, which
+            # saves reading knit data over the wire when we know there are
+            # mising records.
+            text_nodes = set(text_nodes)
+            present_text_keys = set(_node[1] for _node in text_nodes)
+            missing_text_keys = set(text_filter) - present_text_keys
+            if missing_text_keys:
+                # TODO: raise a specific error that can handle many missing
+                # keys.
+                a_missing_key = missing_text_keys.pop()
+                raise errors.RevisionNotPresent(a_missing_key[1],
+                    a_missing_key[0])
+        # copy text keys and adjust values
+        list(self._copy_nodes_graph(text_nodes, text_index_map,
+            new_pack._writer, new_pack.text_index))
+        if 'pack' in debug.debug_flags:
+            mutter('%s: create_pack: file texts copied: %s%s %d items t+%6.3fs',
+                time.ctime(), self._upload_transport.base, new_pack.random_name,
+                new_pack.text_index.key_count(),
+                time.time() - new_pack.start_time)
+        # select signature keys
+        signature_filter = revision_keys # same keyspace
+        signature_index_map = self._packs_list_to_pack_map_and_index_list(
+            packs, 'signature_index')[0]
+        signature_nodes = self._index_contents(signature_index_map,
+            signature_filter)
+        # copy signature keys and adjust values
+        self._copy_nodes(signature_nodes, signature_index_map, new_pack._writer,
+            new_pack.signature_index)
+        if 'pack' in debug.debug_flags:
+            mutter('%s: create_pack: revision signatures copied: %s%s %d items t+%6.3fs',
+                time.ctime(), self._upload_transport.base, new_pack.random_name,
+                new_pack.signature_index.key_count(),
+                time.time() - new_pack.start_time)
+        if not new_pack.data_inserted():
+            new_pack.abort()
+            return None
+        new_pack.finish()
+        self.allocate(new_pack)
+        return new_pack
+
+    def _execute_pack_operations(self, pack_operations):
+        """Execute a series of pack operations.
+
+        :param pack_operations: A list of [revision_count, packs_to_combine].
+        :return: None.
+        """
+        for revision_count, packs in pack_operations:
+            # we may have no-ops from the setup logic
+            if len(packs) == 0:
+                continue
+            # have a progress bar?
+            self.create_pack_from_packs(packs, '.autopack')
+            for pack in packs:
+                self._remove_pack_from_memory(pack)
+        # record the newly available packs and stop advertising the old
+        # packs
+        self._save_pack_names()
+        # Move the old packs out of the way now they are no longer referenced.
+        for revision_count, packs in pack_operations:
+            self._obsolete_packs(packs)
+
+    def lock_names(self):
+        """Acquire the mutex around the pack-names index.
+        
+        This cannot be used in the middle of a read-only transaction on the
+        repository.
+        """
+        self.repo.control_files.lock_write()
+
+    def pack(self):
+        """Pack the pack collection totally."""
+        self.ensure_loaded()
+        total_packs = len(self._names)
+        if total_packs < 2:
+            return
+        total_revisions = self.revision_index.combined_index.key_count()
+        # XXX: the following may want to be a class, to pack with a given
+        # policy.
+        mutter('Packing repository %s, which has %d pack files, '
+            'containing %d revisions into 1 packs.', self, total_packs,
+            total_revisions)
+        # determine which packs need changing
+        pack_distribution = [1]
+        pack_operations = [[0, []]]
+        for pack in self.all_packs():
+            revision_count = pack.get_revision_count()
+            pack_operations[-1][0] += revision_count
+            pack_operations[-1][1].append(pack)
+        self._execute_pack_operations(pack_operations)
+
+    def plan_autopack_combinations(self, existing_packs, pack_distribution):
+        """Plan a pack operation.
+
+        :param existing_packs: The packs to pack. (A list of (revcount, Pack)
+            tuples).
+        :param pack_distribution: A list with the number of revisions desired
+            in each pack.
+        """
+        if len(existing_packs) <= len(pack_distribution):
+            return []
+        existing_packs.sort(reverse=True)
+        pack_operations = [[0, []]]
+        # plan out what packs to keep, and what to reorganise
+        while len(existing_packs):
+            # take the largest pack, and if its less than the head of the
+            # distribution chart we will include its contents in the new pack for
+            # that position. If its larger, we remove its size from the
+            # distribution chart
+            next_pack_rev_count, next_pack = existing_packs.pop(0)
+            if next_pack_rev_count >= pack_distribution[0]:
+                # this is already packed 'better' than this, so we can
+                # not waste time packing it.
+                while next_pack_rev_count > 0:
+                    next_pack_rev_count -= pack_distribution[0]
+                    if next_pack_rev_count >= 0:
+                        # more to go
+                        del pack_distribution[0]
+                    else:
+                        # didn't use that entire bucket up
+                        pack_distribution[0] = -next_pack_rev_count
+            else:
+                # add the revisions we're going to add to the next output pack
+                pack_operations[-1][0] += next_pack_rev_count
+                # allocate this pack to the next pack sub operation
+                pack_operations[-1][1].append(next_pack)
+                if pack_operations[-1][0] >= pack_distribution[0]:
+                    # this pack is used up, shift left.
+                    del pack_distribution[0]
+                    pack_operations.append([0, []])
+        
+        return pack_operations
+
+    def _copy_nodes(self, nodes, index_map, writer, write_index):
+        # plan a readv on each source pack:
+        # group by pack
+        nodes = sorted(nodes)
+        # how to map this into knit.py - or knit.py into this?
+        # we don't want the typical knit logic, we want grouping by pack
+        # at this point - perhaps a helper library for the following code 
+        # duplication points?
+        request_groups = {}
+        for index, key, value in nodes:
+            if index not in request_groups:
+                request_groups[index] = []
+            request_groups[index].append((key, value))
+        for index, items in request_groups.iteritems():
+            pack_readv_requests = []
+            for key, value in items:
+                # ---- KnitGraphIndex.get_position
+                bits = value[1:].split(' ')
+                offset, length = int(bits[0]), int(bits[1])
+                pack_readv_requests.append((offset, length, (key, value[0])))
+            # linear scan up the pack
+            pack_readv_requests.sort()
+            # copy the data
+            transport, path = index_map[index]
+            reader = pack.make_readv_reader(transport, path,
+                [offset[0:2] for offset in pack_readv_requests])
+            for (names, read_func), (_1, _2, (key, eol_flag)) in \
+                izip(reader.iter_records(), pack_readv_requests):
+                raw_data = read_func(None)
+                pos, size = writer.add_bytes_record(raw_data, names)
+                write_index.add_node(key, eol_flag + "%d %d" % (pos, size))
+
+    def _copy_nodes_graph(self, nodes, index_map, writer, write_index,
+        output_lines=False):
+        """Copy knit nodes between packs.
+
+        :param output_lines: Return lines present in the copied data as
+            an iterator.
+        """
+        # for record verification
+        knit_data = _KnitData(None)
+        # for line extraction when requested (inventories only)
+        if output_lines:
+            factory = knit.KnitPlainFactory()
+        # plan a readv on each source pack:
+        # group by pack
+        nodes = sorted(nodes)
+        # how to map this into knit.py - or knit.py into this?
+        # we don't want the typical knit logic, we want grouping by pack
+        # at this point - perhaps a helper library for the following code 
+        # duplication points?
+        request_groups = {}
+        for index, key, value, references in nodes:
+            if index not in request_groups:
+                request_groups[index] = []
+            request_groups[index].append((key, value, references))
+        for index, items in request_groups.iteritems():
+            pack_readv_requests = []
+            for key, value, references in items:
+                # ---- KnitGraphIndex.get_position
+                bits = value[1:].split(' ')
+                offset, length = int(bits[0]), int(bits[1])
+                pack_readv_requests.append((offset, length, (key, value[0], references)))
+            # linear scan up the pack
+            pack_readv_requests.sort()
+            # copy the data
+            transport, path = index_map[index]
+            reader = pack.make_readv_reader(transport, path,
+                [offset[0:2] for offset in pack_readv_requests])
+            for (names, read_func), (_1, _2, (key, eol_flag, references)) in \
+                izip(reader.iter_records(), pack_readv_requests):
+                raw_data = read_func(None)
+                if output_lines:
+                    # read the entire thing
+                    content, _ = knit_data._parse_record(key[-1], raw_data)
+                    if len(references[-1]) == 0:
+                        line_iterator = factory.get_fulltext_content(content)
+                    else:
+                        line_iterator = factory.get_linedelta_content(content)
+                    for line in line_iterator:
+                        yield line
+                else:
+                    # check the header only
+                    df, _ = knit_data._parse_record_header(key[-1], raw_data)
+                    df.close()
+                pos, size = writer.add_bytes_record(raw_data, names)
+                write_index.add_node(key, eol_flag + "%d %d" % (pos, size), references)
+
+    def ensure_loaded(self):
+        # NB: if you see an assertion error here, its probably access against
+        # an unlocked repo. Naughty.
+        assert self.repo.is_locked()
+        if self._names is None:
+            self._names = {}
+            self._packs_at_load = set()
+            for index, key, value in self._iter_disk_pack_index():
+                name = key[0]
+                self._names[name] = self._parse_index_sizes(value)
+                self._packs_at_load.add((key, value))
+        # populate all the metadata.
+        self.all_packs()
+
+    def _parse_index_sizes(self, value):
+        """Parse a string of index sizes."""
+        return tuple([int(digits) for digits in value.split(' ')])
+
+    def get_pack_by_name(self, name):
+        """Get a Pack object by name.
+
+        :param name: The name of the pack - e.g. '123456'
+        :return: A Pack object.
+        """
+        try:
+            return self._packs_by_name[name]
+        except KeyError:
+            rev_index = self._make_index(name, '.rix')
+            inv_index = self._make_index(name, '.iix')
+            txt_index = self._make_index(name, '.tix')
+            sig_index = self._make_index(name, '.six')
+            result = ExistingPack(self._pack_transport, name, rev_index,
+                inv_index, txt_index, sig_index)
+            self.add_pack_to_memory(result)
+            return result
+
+    def allocate(self, a_new_pack):
+        """Allocate name in the list of packs.
+
+        :param a_new_pack: A NewPack instance to be added to the collection of
+            packs for this repository.
+        """
+        self.ensure_loaded()
+        if a_new_pack.name in self._names:
+            # a collision with the packs we know about (not the only possible
+            # collision, see NewPack.finish() for some discussion). Remove our
+            # prior reference to it.
+            self._remove_pack_from_memory(a_new_pack)
+        self._names[a_new_pack.name] = tuple(a_new_pack.index_sizes)
+        self.add_pack_to_memory(a_new_pack)
+
+    def _iter_disk_pack_index(self):
+        """Iterate over the contents of the pack-names index.
+        
+        This is used when loading the list from disk, and before writing to
+        detect updates from others during our write operation.
+        :return: An iterator of the index contents.
+        """
+        return GraphIndex(self.transport, 'pack-names', None
+                ).iter_all_entries()
+
+    def _make_index(self, name, suffix):
+        size_offset = self._suffix_offsets[suffix]
+        index_name = name + suffix
+        index_size = self._names[name][size_offset]
+        return GraphIndex(
+            self._index_transport, index_name, index_size)
+
+    def _max_pack_count(self, total_revisions):
+        """Return the maximum number of packs to use for total revisions.
+        
+        :param total_revisions: The total number of revisions in the
+            repository.
+        """
+        if not total_revisions:
+            return 1
+        digits = str(total_revisions)
+        result = 0
+        for digit in digits:
+            result += int(digit)
+        return result
+
+    def names(self):
+        """Provide an order to the underlying names."""
+        return sorted(self._names.keys())
+
+    def _obsolete_packs(self, packs):
+        """Move a number of packs which have been obsoleted out of the way.
+
+        Each pack and its associated indices are moved out of the way.
+
+        Note: for correctness this function should only be called after a new
+        pack names index has been written without these pack names, and with
+        the names of packs that contain the data previously available via these
+        packs.
+
+        :param packs: The packs to obsolete.
+        :param return: None.
+        """
+        for pack in packs:
+            pack.pack_transport.rename(pack.file_name(),
+                '../obsolete_packs/' + pack.file_name())
+            # TODO: Probably needs to know all possible indices for this pack
+            # - or maybe list the directory and move all indices matching this
+            # name whether we recognize it or not?
+            for suffix in ('.iix', '.six', '.tix', '.rix'):
+                self._index_transport.rename(pack.name + suffix,
+                    '../obsolete_packs/' + pack.name + suffix)
+
+    def pack_distribution(self, total_revisions):
+        """Generate a list of the number of revisions to put in each pack.
+
+        :param total_revisions: The total number of revisions in the
+            repository.
+        """
+        if total_revisions == 0:
+            return [0]
+        digits = reversed(str(total_revisions))
+        result = []
+        for exponent, count in enumerate(digits):
+            size = 10 ** exponent
+            for pos in range(int(count)):
+                result.append(size)
+        return list(reversed(result))
+
+    def _pack_tuple(self, name):
+        """Return a tuple with the transport and file name for a pack name."""
+        return self._pack_transport, name + '.pack'
+
+    def _remove_pack_from_memory(self, pack):
+        """Remove pack from the packs accessed by this repository.
+        
+        Only affects memory state, until self._save_pack_names() is invoked.
+        """
+        self._names.pop(pack.name)
+        self._packs_by_name.pop(pack.name)
+        self._remove_pack_indices(pack)
+
+    def _remove_pack_indices(self, pack):
+        """Remove the indices for pack from the aggregated indices."""
+        self.revision_index.remove_index(pack.revision_index, pack)
+        self.inventory_index.remove_index(pack.inventory_index, pack)
+        self.text_index.remove_index(pack.text_index, pack)
+        self.signature_index.remove_index(pack.signature_index, pack)
+
+    def reset(self):
+        """Clear all cached data."""
+        # cached revision data
+        self.repo._revision_knit = None
+        self.revision_index.clear()
+        # cached signature data
+        self.repo._signature_knit = None
+        self.signature_index.clear()
+        # cached file text data
+        self.text_index.clear()
+        self.repo._text_knit = None
+        # cached inventory data
+        self.inventory_index.clear()
+        # remove the open pack
+        self._new_pack = None
+        # information about packs.
+        self._names = None
+        self.packs = []
+        self._packs_by_name = {}
+        self._packs_at_load = None
+
+    def _make_index_map(self, index_suffix):
+        """Return information on existing indices.
+
+        :param suffix: Index suffix added to pack name.
+
+        :returns: (pack_map, indices) where indices is a list of GraphIndex 
+        objects, and pack_map is a mapping from those objects to the 
+        pack tuple they describe.
+        """
+        # TODO: stop using this; it creates new indices unnecessarily.
+        self.ensure_loaded()
+        suffix_map = {'.rix': 'revision_index',
+            '.six': 'signature_index',
+            '.iix': 'inventory_index',
+            '.tix': 'text_index',
+        }
+        return self._packs_list_to_pack_map_and_index_list(self.all_packs(),
+            suffix_map[index_suffix])
+
+    def _packs_list_to_pack_map_and_index_list(self, packs, index_attribute):
+        """Convert a list of packs to an index pack map and index list.
+
+        :param packs: The packs list to process.
+        :param index_attribute: The attribute that the desired index is found
+            on.
+        :return: A tuple (map, list) where map contains the dict from
+            index:pack_tuple, and lsit contains the indices in the same order
+            as the packs list.
+        """
+        indices = []
+        pack_map = {}
+        for pack in packs:
+            index = getattr(pack, index_attribute)
+            indices.append(index)
+            pack_map[index] = (pack.pack_transport, pack.file_name())
+        return pack_map, indices
+
+    def _index_contents(self, pack_map, key_filter=None):
+        """Get an iterable of the index contents from a pack_map.
+
+        :param pack_map: A map from indices to pack details.
+        :param key_filter: An optional filter to limit the
+            keys returned.
+        """
+        indices = [index for index in pack_map.iterkeys()]
+        all_index = CombinedGraphIndex(indices)
+        if key_filter is None:
+            return all_index.iter_all_entries()
+        else:
+            return all_index.iter_entries(key_filter)
+
+    def _unlock_names(self):
+        """Release the mutex around the pack-names index."""
+        self.repo.control_files.unlock()
+
+    def _save_pack_names(self):
+        """Save the list of packs.
+
+        This will take out the mutex around the pack names list for the
+        duration of the method call. If concurrent updates have been made, a
+        three-way merge between the current list and the current in memory list
+        is performed.
+        """
+        self.lock_names()
+        try:
+            builder = GraphIndexBuilder()
+            # load the disk nodes across
+            disk_nodes = set()
+            for index, key, value in self._iter_disk_pack_index():
+                disk_nodes.add((key, value))
+            # do a two-way diff against our original content
+            current_nodes = set()
+            for name, sizes in self._names.iteritems():
+                current_nodes.add(
+                    ((name, ), ' '.join(str(size) for size in sizes)))
+            deleted_nodes = self._packs_at_load - current_nodes
+            new_nodes = current_nodes - self._packs_at_load
+            disk_nodes.difference_update(deleted_nodes)
+            disk_nodes.update(new_nodes)
+            # TODO: handle same-name, index-size-changes here - 
+            # e.g. use the value from disk, not ours, *unless* we're the one
+            # changing it.
+            for key, value in disk_nodes:
+                builder.add_node(key, value)
+            self.transport.put_file('pack-names', builder.finish())
+            # move the baseline forward
+            self._packs_at_load = disk_nodes
+        finally:
+            self._unlock_names()
+        # synchronise the memory packs list with what we just wrote:
+        new_names = dict(disk_nodes)
+        # drop no longer present nodes
+        for pack in self.all_packs():
+            if (pack.name,) not in new_names:
+                self._remove_pack_from_memory(pack)
+        # add new nodes/refresh existing ones
+        for key, value in disk_nodes:
+            name = key[0]
+            sizes = self._parse_index_sizes(value)
+            if name in self._names:
+                # existing
+                if sizes != self._names[name]:
+                    # the pack for name has had its indices replaced - rare but
+                    # important to handle. XXX: probably can never happen today
+                    # because the three-way merge code above does not handle it
+                    # - you may end up adding the same key twice to the new
+                    # disk index because the set values are the same, unless
+                    # the only index shows up as deleted by the set difference
+                    # - which it may. Until there is a specific test for this,
+                    # assume its broken. RBC 20071017.
+                    self._remove_pack_from_memory(self.get_pack_by_name(name))
+                    self._names[name] = sizes
+                    self.get_pack_by_name(name)
+            else:
+                # new
+                self._names[name] = sizes
+                self.get_pack_by_name(name)
+
+    def _start_write_group(self):
+        # Do not permit preparation for writing if we're not in a 'write lock'.
+        if not self.repo.is_write_locked():
+            raise errors.NotWriteLocked(self)
+        self._new_pack = NewPack(self._upload_transport, self._index_transport,
+            self._pack_transport, upload_suffix='.pack')
+        # allow writing: queue writes to a new index
+        self.revision_index.add_writable_index(self._new_pack.revision_index,
+            self._new_pack)
+        self.inventory_index.add_writable_index(self._new_pack.inventory_index,
+            self._new_pack)
+        self.text_index.add_writable_index(self._new_pack.text_index,
+            self._new_pack)
+        self.signature_index.add_writable_index(self._new_pack.signature_index,
+            self._new_pack)
+
+        # reused revision and signature knits may need updating
+        #
+        # "Hysterical raisins. client code in bzrlib grabs those knits outside
+        # of write groups and then mutates it inside the write group."
+        if self.repo._revision_knit is not None:
+            self.repo._revision_knit._index._add_callback = \
+                self.revision_index.add_callback
+        if self.repo._signature_knit is not None:
+            self.repo._signature_knit._index._add_callback = \
+                self.signature_index.add_callback
+        # create a reused knit object for text addition in commit.
+        self.repo._text_knit = self.repo.weave_store.get_weave_or_empty(
+            'all-texts', None)
+
+    def _abort_write_group(self):
+        # FIXME: just drop the transient index.
+        # forget what names there are
+        self._new_pack.abort()
+        self._remove_pack_indices(self._new_pack)
+        self._new_pack = None
+        self.repo._text_knit = None
+
+    def _commit_write_group(self):
+        self._remove_pack_indices(self._new_pack)
+        if self._new_pack.data_inserted():
+            # get all the data to disk and read to use
+            self._new_pack.finish()
+            self.allocate(self._new_pack)
+            self._new_pack = None
+            if not self.autopack():
+                # when autopack takes no steps, the names list is still
+                # unsaved.
+                self._save_pack_names()
+        else:
+            self._new_pack.abort()
+        self.repo._text_knit = None
+
+
+class KnitPackRevisionStore(KnitRevisionStore):
+    """An object to adapt access from RevisionStore's to use KnitPacks.
+
+    This class works by replacing the original RevisionStore.
+    We need to do this because the KnitPackRevisionStore is less
+    isolated in its layering - it uses services from the repo.
+    """
+
+    def __init__(self, repo, transport, revisionstore):
+        """Create a KnitPackRevisionStore on repo with revisionstore.
+
+        This will store its state in the Repository, use the
+        indices to provide a KnitGraphIndex,
+        and at the end of transactions write new indices.
+        """
+        KnitRevisionStore.__init__(self, revisionstore.versioned_file_store)
+        self.repo = repo
+        self._serializer = revisionstore._serializer
+        self.transport = transport
+
+    def get_revision_file(self, transaction):
+        """Get the revision versioned file object."""
+        if getattr(self.repo, '_revision_knit', None) is not None:
+            return self.repo._revision_knit
+        self.repo._pack_collection.ensure_loaded()
+        add_callback = self.repo._pack_collection.revision_index.add_callback
+        # setup knit specific objects
+        knit_index = KnitGraphIndex(
+            self.repo._pack_collection.revision_index.combined_index,
+            add_callback=add_callback)
+        self.repo._revision_knit = knit.KnitVersionedFile(
+            'revisions', self.transport.clone('..'),
+            self.repo.control_files._file_mode,
+            create=False, access_mode=self.repo._access_mode(),
+            index=knit_index, delta=False, factory=knit.KnitPlainFactory(),
+            access_method=self.repo._pack_collection.revision_index.knit_access)
+        return self.repo._revision_knit
+
+    def get_signature_file(self, transaction):
+        """Get the signature versioned file object."""
+        if getattr(self.repo, '_signature_knit', None) is not None:
+            return self.repo._signature_knit
+        self.repo._pack_collection.ensure_loaded()
+        add_callback = self.repo._pack_collection.signature_index.add_callback
+        # setup knit specific objects
+        knit_index = KnitGraphIndex(
+            self.repo._pack_collection.signature_index.combined_index,
+            add_callback=add_callback, parents=False)
+        self.repo._signature_knit = knit.KnitVersionedFile(
+            'signatures', self.transport.clone('..'),
+            self.repo.control_files._file_mode,
+            create=False, access_mode=self.repo._access_mode(),
+            index=knit_index, delta=False, factory=knit.KnitPlainFactory(),
+            access_method=self.repo._pack_collection.signature_index.knit_access)
+        return self.repo._signature_knit
+
+
+class KnitPackTextStore(VersionedFileStore):
+    """Presents a TextStore abstraction on top of packs.
+
+    This class works by replacing the original VersionedFileStore.
+    We need to do this because the KnitPackRevisionStore is less
+    isolated in its layering - it uses services from the repo and shares them
+    with all the data written in a single write group.
+    """
+
+    def __init__(self, repo, transport, weavestore):
+        """Create a KnitPackTextStore on repo with weavestore.
+
+        This will store its state in the Repository, use the
+        indices FileNames to provide a KnitGraphIndex,
+        and at the end of transactions write new indices.
+        """
+        # don't call base class constructor - it's not suitable.
+        # no transient data stored in the transaction
+        # cache.
+        self._precious = False
+        self.repo = repo
+        self.transport = transport
+        self.weavestore = weavestore
+        # XXX for check() which isn't updated yet
+        self._transport = weavestore._transport
+
+    def get_weave_or_empty(self, file_id, transaction):
+        """Get a 'Knit' backed by the .tix indices.
+
+        The transaction parameter is ignored.
+        """
+        self.repo._pack_collection.ensure_loaded()
+        add_callback = self.repo._pack_collection.text_index.add_callback
+        # setup knit specific objects
+        file_id_index = GraphIndexPrefixAdapter(
+            self.repo._pack_collection.text_index.combined_index,
+            (file_id, ), 1, add_nodes_callback=add_callback)
+        knit_index = KnitGraphIndex(file_id_index,
+            add_callback=file_id_index.add_nodes,
+            deltas=True, parents=True)
+        return knit.KnitVersionedFile('text:' + file_id,
+            self.transport.clone('..'),
+            None,
+            index=knit_index,
+            access_method=self.repo._pack_collection.text_index.knit_access,
+            factory=knit.KnitPlainFactory())
+
+    get_weave = get_weave_or_empty
+
+    def __iter__(self):
+        """Generate a list of the fileids inserted, for use by check."""
+        self.repo._pack_collection.ensure_loaded()
+        ids = set()
+        for index, key, value, refs in \
+            self.repo._pack_collection.text_index.combined_index.iter_all_entries():
+            ids.add(key[0])
+        return iter(ids)
+
+
+class InventoryKnitThunk(object):
+    """An object to manage thunking get_inventory_weave to pack based knits."""
+
+    def __init__(self, repo, transport):
+        """Create an InventoryKnitThunk for repo at transport.
+
+        This will store its state in the Repository, use the
+        indices FileNames to provide a KnitGraphIndex,
+        and at the end of transactions write a new index..
+        """
+        self.repo = repo
+        self.transport = transport
+
+    def get_weave(self):
+        """Get a 'Knit' that contains inventory data."""
+        self.repo._pack_collection.ensure_loaded()
+        add_callback = self.repo._pack_collection.inventory_index.add_callback
+        # setup knit specific objects
+        knit_index = KnitGraphIndex(
+            self.repo._pack_collection.inventory_index.combined_index,
+            add_callback=add_callback, deltas=True, parents=True)
+        return knit.KnitVersionedFile(
+            'inventory', self.transport.clone('..'),
+            self.repo.control_files._file_mode,
+            create=False, access_mode=self.repo._access_mode(),
+            index=knit_index, delta=True, factory=knit.KnitPlainFactory(),
+            access_method=self.repo._pack_collection.inventory_index.knit_access)
+
+
+class KnitPackRepository(KnitRepository):
+    """Experimental graph-knit using repository."""
+
+    def __init__(self, _format, a_bzrdir, control_files, _revision_store,
+        control_store, text_store, _commit_builder_class, _serializer):
+        KnitRepository.__init__(self, _format, a_bzrdir, control_files,
+            _revision_store, control_store, text_store, _commit_builder_class,
+            _serializer)
+        index_transport = control_files._transport.clone('indices')
+        self._pack_collection = RepositoryPackCollection(self, control_files._transport,
+            index_transport,
+            control_files._transport.clone('upload'),
+            control_files._transport.clone('packs'))
+        self._revision_store = KnitPackRevisionStore(self, index_transport, self._revision_store)
+        self.weave_store = KnitPackTextStore(self, index_transport, self.weave_store)
+        self._inv_thunk = InventoryKnitThunk(self, index_transport)
+        # True when the repository object is 'write locked' (as opposed to the
+        # physical lock only taken out around changes to the pack-names list.) 
+        # Another way to represent this would be a decorator around the control
+        # files object that presents logical locks as physical ones - if this
+        # gets ugly consider that alternative design. RBC 20071011
+        self._write_lock_count = 0
+        self._transaction = None
+        # for tests
+        self._reconcile_does_inventory_gc = False
+        self._reconcile_fixes_text_parents = False
+
+    def _abort_write_group(self):
+        self._pack_collection._abort_write_group()
+
+    def _access_mode(self):
+        """Return 'w' or 'r' for depending on whether a write lock is active.
+        
+        This method is a helper for the Knit-thunking support objects.
+        """
+        if self.is_write_locked():
+            return 'w'
+        return 'r'
+
+    def get_parents(self, revision_ids):
+        """See StackedParentsProvider.get_parents.
+        
+        This implementation accesses the combined revision index to provide
+        answers.
+        """
+        index = self._pack_collection.revision_index.combined_index
+        search_keys = set()
+        for revision_id in revision_ids:
+            if revision_id != _mod_revision.NULL_REVISION:
+                search_keys.add((revision_id,))
+        found_parents = {_mod_revision.NULL_REVISION:[]}
+        for index, key, value, refs in index.iter_entries(search_keys):
+            parents = refs[0]
+            if not parents:
+                parents = (_mod_revision.NULL_REVISION,)
+            else:
+                parents = tuple(parent[0] for parent in parents)
+            found_parents[key[0]] = parents
+        result = []
+        for revision_id in revision_ids:
+            try:
+                result.append(found_parents[revision_id])
+            except KeyError:
+                result.append(None)
+        return result
+
+    def _make_parents_provider(self):
+        return self
+
+    def _refresh_data(self):
+        if self._write_lock_count == 1 or self.control_files._lock_count == 1:
+            # forget what names there are
+            self._pack_collection.reset()
+            # XXX: Better to do an in-memory merge when acquiring a new lock -
+            # factor out code from _save_pack_names.
+
+    def _start_write_group(self):
+        self._pack_collection._start_write_group()
+
+    def _commit_write_group(self):
+        return self._pack_collection._commit_write_group()
+
+    def get_inventory_weave(self):
+        return self._inv_thunk.get_weave()
+
+    def get_transaction(self):
+        if self._write_lock_count:
+            return self._transaction
+        else:
+            return self.control_files.get_transaction()
+
+    def is_locked(self):
+        return self._write_lock_count or self.control_files.is_locked()
+
+    def is_write_locked(self):
+        return self._write_lock_count
+
+    def lock_write(self, token=None):
+        if not self._write_lock_count and self.is_locked():
+            raise errors.ReadOnlyError(self)
+        self._write_lock_count += 1
+        if self._write_lock_count == 1:
+            from bzrlib import transactions
+            self._transaction = transactions.WriteTransaction()
+        self._refresh_data()
+
+    def lock_read(self):
+        if self._write_lock_count:
+            self._write_lock_count += 1
+        else:
+            self.control_files.lock_read()
+        self._refresh_data()
+
+    def leave_lock_in_place(self):
+        # not supported - raise an error
+        raise NotImplementedError(self.leave_lock_in_place)
+
+    def dont_leave_lock_in_place(self):
+        # not supported - raise an error
+        raise NotImplementedError(self.dont_leave_lock_in_place)
+
+    @needs_write_lock
+    def pack(self):
+        """Compress the data within the repository.
+
+        This will pack all the data to a single pack. In future it may
+        recompress deltas or do other such expensive operations.
+        """
+        self._pack_collection.pack()
+
+    @needs_write_lock
+    def reconcile(self, other=None, thorough=False):
+        """Reconcile this repository."""
+        from bzrlib.reconcile import PackReconciler
+        reconciler = PackReconciler(self, thorough=thorough)
+        reconciler.reconcile()
+        return reconciler
+
+    def unlock(self):
+        if self._write_lock_count == 1 and self._write_group is not None:
+            raise errors.BzrError(
+                'Must end write groups before releasing write locks.')
+        if self._write_lock_count:
+            self._write_lock_count -= 1
+            if not self._write_lock_count:
+                transaction = self._transaction
+                self._transaction = None
+                transaction.finish()
+        else:
+            self.control_files.unlock()
+
+
+class RepositoryFormatPack(MetaDirRepositoryFormat):
+    """Format logic for pack structured repositories.
+
+    This repository format has:
+     - a list of packs in pack-names
+     - packs in packs/NAME.pack
+     - indices in indices/NAME.{iix,six,tix,rix}
+     - knit deltas in the packs, knit indices mapped to the indices.
+     - thunk objects to support the knits programming API.
+     - a format marker of its own
+     - an optional 'shared-storage' flag
+     - an optional 'no-working-trees' flag
+     - a LockDir lock
+    """
+
+    # Set this attribute in derived classes to control the repository class
+    # created by open and initialize.
+    repository_class = None
+    # Set this attribute in derived classes to control the
+    # _commit_builder_class that the repository objects will have passed to
+    # their constructor.
+    _commit_builder_class = None
+    # Set this attribute in derived clases to control the _serializer that the
+    # repository objects will have passed to their constructor.
+    _serializer = None
+
+    def _get_control_store(self, repo_transport, control_files):
+        """Return the control store for this repository."""
+        return VersionedFileStore(
+            repo_transport,
+            prefixed=False,
+            file_mode=control_files._file_mode,
+            versionedfile_class=knit.KnitVersionedFile,
+            versionedfile_kwargs={'factory': knit.KnitPlainFactory()},
+            )
+
+    def _get_revision_store(self, repo_transport, control_files):
+        """See RepositoryFormat._get_revision_store()."""
+        versioned_file_store = VersionedFileStore(
+            repo_transport,
+            file_mode=control_files._file_mode,
+            prefixed=False,
+            precious=True,
+            versionedfile_class=knit.KnitVersionedFile,
+            versionedfile_kwargs={'delta': False,
+                                  'factory': knit.KnitPlainFactory(),
+                                 },
+            escaped=True,
+            )
+        return KnitRevisionStore(versioned_file_store)
+
+    def _get_text_store(self, transport, control_files):
+        """See RepositoryFormat._get_text_store()."""
+        return self._get_versioned_file_store('knits',
+                                  transport,
+                                  control_files,
+                                  versionedfile_class=knit.KnitVersionedFile,
+                                  versionedfile_kwargs={
+                                      'create_parent_dir': True,
+                                      'delay_create': True,
+                                      'dir_mode': control_files._dir_mode,
+                                  },
+                                  escaped=True)
+
+    def initialize(self, a_bzrdir, shared=False):
+        """Create a pack based repository.
+
+        :param a_bzrdir: bzrdir to contain the new repository; must already
+            be initialized.
+        :param shared: If true the repository will be initialized as a shared
+                       repository.
+        """
+        mutter('creating repository in %s.', a_bzrdir.transport.base)
+        dirs = ['indices', 'obsolete_packs', 'packs', 'upload']
+        builder = GraphIndexBuilder()
+        files = [('pack-names', builder.finish())]
+        utf8_files = [('format', self.get_format_string())]
+        
+        self._upload_blank_content(a_bzrdir, dirs, files, utf8_files, shared)
+        return self.open(a_bzrdir=a_bzrdir, _found=True)
+
+    def open(self, a_bzrdir, _found=False, _override_transport=None):
+        """See RepositoryFormat.open().
+        
+        :param _override_transport: INTERNAL USE ONLY. Allows opening the
+                                    repository at a slightly different url
+                                    than normal. I.e. during 'upgrade'.
+        """
+        if not _found:
+            format = RepositoryFormat.find_format(a_bzrdir)
+            assert format.__class__ ==  self.__class__
+        if _override_transport is not None:
+            repo_transport = _override_transport
+        else:
+            repo_transport = a_bzrdir.get_repository_transport(None)
+        control_files = lockable_files.LockableFiles(repo_transport,
+                                'lock', lockdir.LockDir)
+        text_store = self._get_text_store(repo_transport, control_files)
+        control_store = self._get_control_store(repo_transport, control_files)
+        _revision_store = self._get_revision_store(repo_transport, control_files)
+        return self.repository_class(_format=self,
+                              a_bzrdir=a_bzrdir,
+                              control_files=control_files,
+                              _revision_store=_revision_store,
+                              control_store=control_store,
+                              text_store=text_store,
+                              _commit_builder_class=self._commit_builder_class,
+                              _serializer=self._serializer)
+
+
+class RepositoryFormatKnitPack1(RepositoryFormatPack):
+    """A no-subtrees parameterised Pack repository.
+
+    This format was introduced in bzr.dev.
+    """
+
+    repository_class = KnitPackRepository
+    _commit_builder_class = PackCommitBuilder
+    _serializer = xml5.serializer_v5
+
+    def _get_matching_bzrdir(self):
+        return bzrdir.format_registry.make_bzrdir('experimental')
+
+    def _ignore_setting_bzrdir(self, format):
+        pass
+
+    _matchingbzrdir = property(_get_matching_bzrdir, _ignore_setting_bzrdir)
+
+    def get_format_string(self):
+        """See RepositoryFormat.get_format_string()."""
+        return "Bazaar Experimental no-subtrees\n"
+
+    def get_format_description(self):
+        """See RepositoryFormat.get_format_description()."""
+        return "Experimental no-subtrees"
+
+    def check_conversion_target(self, target_format):
+        pass
+
+
+class RepositoryFormatKnitPack3(RepositoryFormatPack):
+    """A subtrees parameterised Pack repository.
+
+    This repository format uses the xml7 serializer to get:
+     - support for recording full info about the tree root
+     - support for recording tree-references
+
+    This format was introduced in bzr.dev.
+    """
+
+    repository_class = KnitPackRepository
+    _commit_builder_class = PackRootCommitBuilder
+    rich_root_data = True
+    supports_tree_reference = True
+    _serializer = xml7.serializer_v7
+
+    def _get_matching_bzrdir(self):
+        return bzrdir.format_registry.make_bzrdir('experimental-subtree')
+
+    def _ignore_setting_bzrdir(self, format):
+        pass
+
+    _matchingbzrdir = property(_get_matching_bzrdir, _ignore_setting_bzrdir)
+
+    def check_conversion_target(self, target_format):
+        if not target_format.rich_root_data:
+            raise errors.BadConversionTarget(
+                'Does not support rich root data.', target_format)
+        if not getattr(target_format, 'supports_tree_reference', False):
+            raise errors.BadConversionTarget(
+                'Does not support nested trees', target_format)
+            
+    def get_format_string(self):
+        """See RepositoryFormat.get_format_string()."""
+        return "Bazaar Experimental subtrees\n"
+
+    def get_format_description(self):
+        """See RepositoryFormat.get_format_description()."""
+        return "Experimental subtrees\n"

=== modified file 'bzrlib/repository.py'
--- bzrlib/repository.py	2007-10-19 17:00:10 +0000
+++ bzrlib/repository.py	2007-10-24 05:59:51 +0000
@@ -494,6 +494,7 @@
 
         returns the sha1 of the serialized inventory.
         """
+        assert self.is_in_write_group()
         _mod_revision.check_not_reserved_id(revision_id)
         assert inv.revision_id is None or inv.revision_id == revision_id, \
             "Mismatch between inventory revision" \
@@ -758,6 +759,11 @@
         raise NotImplementedError(self.get_data_stream)
 
     def insert_data_stream(self, stream):
+        """XXX What does this really do? 
+        
+        Is it a substitute for fetch? 
+        Should it manage its own write group ?
+        """
         for item_key, bytes in stream:
             if item_key[0] == 'file':
                 (file_id,) = item_key[1:]
@@ -1043,19 +1049,18 @@
                                                          signature,
                                                          self.get_transaction())
 
-    def fileids_altered_by_revision_ids(self, revision_ids):
-        """Find the file ids and versions affected by revisions.
-
-        :param revisions: an iterable containing revision ids.
+    def _find_file_ids_from_xml_inventory_lines(self, line_iterator,
+        revision_ids):
+        """Helper routine for fileids_altered_by_revision_ids.
+
+        This performs the translation of xml lines to revision ids.
+
+        :param line_iterator: An iterator of lines
+        :param revision_ids: The revision ids to filter for.
         :return: a dictionary mapping altered file-ids to an iterable of
         revision_ids. Each altered file-ids has the exact revision_ids that
         altered it listed explicitly.
         """
-        assert self._serializer.support_altered_by_hack, \
-            ("fileids_altered_by_revision_ids only supported for branches " 
-             "which store inventory as unnested xml, not on %r" % self)
-        selected_revision_ids = set(revision_ids)
-        w = self.get_inventory_weave()
         result = {}
 
         # this code needs to read every new line in every inventory for the
@@ -1077,44 +1082,61 @@
         search = self._file_ids_altered_regex.search
         unescape = _unescape_xml
         setdefault = result.setdefault
+        for line in line_iterator:
+            match = search(line)
+            if match is None:
+                continue
+            # One call to match.group() returning multiple items is quite a
+            # bit faster than 2 calls to match.group() each returning 1
+            file_id, revision_id = match.group('file_id', 'revision_id')
+
+            # Inlining the cache lookups helps a lot when you make 170,000
+            # lines and 350k ids, versus 8.4 unique ids.
+            # Using a cache helps in 2 ways:
+            #   1) Avoids unnecessary decoding calls
+            #   2) Re-uses cached strings, which helps in future set and
+            #      equality checks.
+            # (2) is enough that removing encoding entirely along with
+            # the cache (so we are using plain strings) results in no
+            # performance improvement.
+            try:
+                revision_id = unescape_revid_cache[revision_id]
+            except KeyError:
+                unescaped = unescape(revision_id)
+                unescape_revid_cache[revision_id] = unescaped
+                revision_id = unescaped
+
+            if revision_id in revision_ids:
+                try:
+                    file_id = unescape_fileid_cache[file_id]
+                except KeyError:
+                    unescaped = unescape(file_id)
+                    unescape_fileid_cache[file_id] = unescaped
+                    file_id = unescaped
+                setdefault(file_id, set()).add(revision_id)
+        return result
+
+    def fileids_altered_by_revision_ids(self, revision_ids):
+        """Find the file ids and versions affected by revisions.
+
+        :param revisions: an iterable containing revision ids.
+        :return: a dictionary mapping altered file-ids to an iterable of
+        revision_ids. Each altered file-ids has the exact revision_ids that
+        altered it listed explicitly.
+        """
+        assert self._serializer.support_altered_by_hack, \
+            ("fileids_altered_by_revision_ids only supported for branches " 
+             "which store inventory as unnested xml, not on %r" % self)
+        selected_revision_ids = set(revision_ids)
+        w = self.get_inventory_weave()
         pb = ui.ui_factory.nested_progress_bar()
         try:
-            for line in w.iter_lines_added_or_present_in_versions(
-                                        selected_revision_ids, pb=pb):
-                match = search(line)
-                if match is None:
-                    continue
-                # One call to match.group() returning multiple items is quite a
-                # bit faster than 2 calls to match.group() each returning 1
-                file_id, revision_id = match.group('file_id', 'revision_id')
-
-                # Inlining the cache lookups helps a lot when you make 170,000
-                # lines and 350k ids, versus 8.4 unique ids.
-                # Using a cache helps in 2 ways:
-                #   1) Avoids unnecessary decoding calls
-                #   2) Re-uses cached strings, which helps in future set and
-                #      equality checks.
-                # (2) is enough that removing encoding entirely along with
-                # the cache (so we are using plain strings) results in no
-                # performance improvement.
-                try:
-                    revision_id = unescape_revid_cache[revision_id]
-                except KeyError:
-                    unescaped = unescape(revision_id)
-                    unescape_revid_cache[revision_id] = unescaped
-                    revision_id = unescaped
-
-                if revision_id in selected_revision_ids:
-                    try:
-                        file_id = unescape_fileid_cache[file_id]
-                    except KeyError:
-                        unescaped = unescape(file_id)
-                        unescape_fileid_cache[file_id] = unescaped
-                        file_id = unescaped
-                    setdefault(file_id, set()).add(revision_id)
+            return self._find_file_ids_from_xml_inventory_lines(
+                w.iter_lines_added_or_present_in_versions(
+                    selected_revision_ids, pb=pb),
+                selected_revision_ids)
         finally:
             pb.finished()
-        return result
 
     def iter_files_bytes(self, desired_files):
         """Iterate through file versions.
@@ -1595,6 +1617,18 @@
 
 def install_revision(repository, rev, revision_tree):
     """Install all revision data into a repository."""
+    repository.start_write_group()
+    try:
+        _install_revision(repository, rev, revision_tree)
+    except:
+        repository.abort_write_group()
+        raise
+    else:
+        repository.commit_write_group()
+
+
+def _install_revision(repository, rev, revision_tree):
+    """Install all revision data into a repository."""
     present_parents = []
     parent_trees = {}
     for p_id in rev.parent_ids:
@@ -1934,6 +1968,20 @@
     'RepositoryFormatKnit3',
     )
 
+# Experimental formats. These make no guarantee about data stability.
+# There is one format for pre-subtrees, and one for post-subtrees to
+# allow ease of testing.
+format_registry.register_lazy(
+    'Bazaar Experimental no-subtrees\n',
+    'bzrlib.repofmt.pack_repo',
+    'RepositoryFormatKnitPack1',
+    )
+format_registry.register_lazy(
+    'Bazaar Experimental subtrees\n',
+    'bzrlib.repofmt.pack_repo',
+    'RepositoryFormatKnitPack3',
+    )
+
 
 class InterRepository(InterObject):
     """This class represents operations taking place between two repositories.
@@ -2244,6 +2292,89 @@
             return self.source._eliminate_revisions_not_present(required_topo_revisions)
 
 
+class InterPackRepo(InterSameDataRepository):
+    """Optimised code paths between Pack based repositories."""
+
+    @classmethod
+    def _get_repo_format_to_test(self):
+        from bzrlib.repofmt import pack_repo
+        return pack_repo.RepositoryFormatKnitPack1()
+
+    @staticmethod
+    def is_compatible(source, target):
+        """Be compatible with known Pack formats.
+        
+        We don't test for the stores being of specific types because that
+        could lead to confusing results, and there is no need to be 
+        overly general.
+        """
+        from bzrlib.repofmt.pack_repo import RepositoryFormatPack
+        try:
+            are_packs = (isinstance(source._format, RepositoryFormatPack) and
+                isinstance(target._format, RepositoryFormatPack))
+        except AttributeError:
+            return False
+        return are_packs and InterRepository._same_model(source, target)
+
+    @needs_write_lock
+    def fetch(self, revision_id=None, pb=None):
+        """See InterRepository.fetch()."""
+        mutter("Using fetch logic to copy between %s(%s) and %s(%s)",
+               self.source, self.source._format, self.target, self.target._format)
+        self.count_copied = 0
+        if revision_id is None:
+            # TODO:
+            # everything to do - use pack logic
+            # to fetch from all packs to one without
+            # inventory parsing etc, IFF nothing to be copied is in the target.
+            # till then:
+            revision_ids = self.source.all_revision_ids()
+            # implementing the TODO will involve:
+            # - detecting when all of a pack is selected
+            # - avoiding as much as possible pre-selection, so the
+            # more-core routines such as create_pack_from_packs can filter in
+            # a just-in-time fashion. (though having a HEADS list on a
+            # repository might make this a lot easier, because we could
+            # sensibly detect 'new revisions' without doing a full index scan.
+        elif _mod_revision.is_null(revision_id):
+            # nothing to do:
+            return
+        else:
+            try:
+                revision_ids = self.missing_revision_ids(revision_id)
+            except errors.NoSuchRevision:
+                raise errors.InstallFailed([revision_id])
+        packs = self.source._pack_collection.all_packs()
+        pack = self.target._pack_collection.create_pack_from_packs(
+            packs, '.fetch', revision_ids,
+            )
+        if pack is not None:
+            self.target._pack_collection._save_pack_names()
+            # Trigger an autopack. This may duplicate effort as we've just done
+            # a pack creation, but for now it is simpler to think about as
+            # 'upload data, then repack if needed'.
+            self.target._pack_collection.autopack()
+            return pack.get_revision_count()
+        else:
+            return 0
+
+    @needs_read_lock
+    def missing_revision_ids(self, revision_id=None):
+        """See InterRepository.missing_revision_ids()."""
+        if revision_id is not None:
+            source_ids = self.source.get_ancestry(revision_id)
+            assert source_ids[0] is None
+            source_ids.pop(0)
+        else:
+            source_ids = self.source.all_revision_ids()
+        # source_ids is the worst possible case we may need to pull.
+        # now we want to filter source_ids against what we actually
+        # have in target, but don't try to check for existence where we know
+        # we do not have a revision as that would be pointless.
+        target_ids = set(self.target.all_revision_ids())
+        return [r for r in source_ids if (r not in target_ids)]
+
+
 class InterModel1and2(InterRepository):
 
     @classmethod
@@ -2299,10 +2430,15 @@
         """Be compatible with Knit1 source and Knit3 target"""
         from bzrlib.repofmt.knitrepo import RepositoryFormatKnit3
         try:
-            from bzrlib.repofmt.knitrepo import RepositoryFormatKnit1, \
-                    RepositoryFormatKnit3
-            return (isinstance(source._format, (RepositoryFormatKnit1)) and
-                    isinstance(target._format, (RepositoryFormatKnit3)))
+            from bzrlib.repofmt.knitrepo import (RepositoryFormatKnit1,
+                RepositoryFormatKnit3)
+            from bzrlib.repofmt.pack_repo import (RepositoryFormatKnitPack1,
+                RepositoryFormatKnitPack3)
+            return (isinstance(source._format,
+                    (RepositoryFormatKnit1, RepositoryFormatKnitPack1)) and
+                isinstance(target._format,
+                    (RepositoryFormatKnit3, RepositoryFormatKnitPack3))
+                )
         except AttributeError:
             return False
 
@@ -2394,6 +2530,7 @@
 InterRepository.register_optimiser(InterKnitRepo)
 InterRepository.register_optimiser(InterModel1and2)
 InterRepository.register_optimiser(InterKnit1and2)
+InterRepository.register_optimiser(InterPackRepo)
 InterRepository.register_optimiser(InterRemoteToOther)
 InterRepository.register_optimiser(InterOtherToRemote)
 
@@ -2563,7 +2700,7 @@
         for parent in parents_from_inventories:
             if parent in heads and parent not in new_parents:
                 new_parents.append(parent)
-        return new_parents
+        return tuple(new_parents)
 
     def check_file_version_parents(self, weave, file_id):
         result = {}
@@ -2574,7 +2711,7 @@
                 continue
             text_revision = self.revision_versions.get_text_version(
                 file_id, revision_id)
-            knit_parents = weave.get_parents(text_revision)
+            knit_parents = tuple(weave.get_parents(text_revision))
             if correct_parents != knit_parents:
                 result[revision_id] = (knit_parents, correct_parents)
         return result

=== modified file 'bzrlib/tests/interrepository_implementations/test_interrepository.py'
--- bzrlib/tests/interrepository_implementations/test_interrepository.py	2007-10-04 01:24:06 +0000
+++ bzrlib/tests/interrepository_implementations/test_interrepository.py	2007-10-17 09:39:41 +0000
@@ -148,6 +148,8 @@
             # check that b now has all the data from a's first commit.
             rev = repo.get_revision('rev1')
             tree = repo.revision_tree('rev1')
+            tree.lock_read()
+            self.addCleanup(tree.unlock)
             tree.get_file_text('file1')
             for file_id in tree:
                 if tree.inventory[file_id].kind == "file":

=== modified file 'bzrlib/tests/repository_implementations/__init__.py'
--- bzrlib/tests/repository_implementations/__init__.py	2007-10-15 05:23:29 +0000
+++ bzrlib/tests/repository_implementations/__init__.py	2007-10-17 09:39:41 +0000
@@ -141,10 +141,10 @@
     """
 
     def all_versions(self):
-        return ['rev1a']
+        return ('rev1a', )
 
     def populated_parents(self):
-        return [([], 'rev1a')]
+        return (((), 'rev1a'), )
 
     def corrected_parents(self):
         # Same as the populated parents, because there was nothing wrong.
@@ -169,23 +169,23 @@
     """
 
     def all_versions(self):
-        return ['rev1a', 'rev1b', 'rev2']
+        return ('rev1a', 'rev1b', 'rev2')
 
     def populated_parents(self):
-        return [
-            ([], 'rev1a'),
-            ([], 'rev1b'),
-            (['rev1a', 'rev1b'], 'rev2')]
+        return (
+            ((), 'rev1a'),
+            ((), 'rev1b'),
+            (('rev1a', 'rev1b'), 'rev2'))
 
     def corrected_parents(self):
-        return [
-            ([], 'rev1a'),
-            ([], 'rev1b'),
-            (['rev1a'], 'rev2')]
+        return (
+            ((), 'rev1a'),
+            ((), 'rev1b'),
+            (('rev1a',), 'rev2'))
 
     def check_regexes(self):
-        return [r"\* a-file-id version rev2 has parents \['rev1a', 'rev1b'\] "
-                r"but should have \['rev1a'\]",
+        return [r"\* a-file-id version rev2 has parents \('rev1a', 'rev1b'\) "
+                r"but should have \('rev1a',\)",
                 "1 unreferenced text ancestors",
                 ]
 
@@ -219,21 +219,21 @@
     """
 
     def all_versions(self):
-        return ['rev2', 'rev3']
+        return ('rev2', 'rev3')
 
     def populated_parents(self):
-        return [
-            ([], 'rev2'),
-            (['rev1c'], 'rev3')]
+        return (
+            ((), 'rev2'),
+            (('rev1c',), 'rev3'))
 
     def corrected_parents(self):
-        return [
-            ([], 'rev2'),
-            ([], 'rev3')]
+        return (
+            ((), 'rev2'),
+            ((), 'rev3'))
 
     def check_regexes(self):
         return [r"\* a-file-id version rev3 has parents "
-                r"\['rev1c'\] but should have \[\]",
+                r"\('rev1c',\) but should have \(\)",
                 # Also check reporting of unreferenced ancestors
                 r"unreferenced ancestor: {rev1c} in a-file-id",
                 ]
@@ -272,35 +272,35 @@
     """
 
     def all_versions(self):
-        return ['rev1a', 'rev2', 'rev4', 'rev2b', 'rev4', 'rev2c', 'rev5']
+        return ('rev1a', 'rev2', 'rev4', 'rev2b', 'rev4', 'rev2c', 'rev5')
 
     def populated_parents(self):
-        return [
-            (['rev2'], 'rev3'),
-            (['rev2'], 'rev4'),
-            (['rev2', 'rev2c'], 'rev5')]
+        return (
+            (('rev2',), 'rev3'),
+            (('rev2',), 'rev4'),
+            (('rev2', 'rev2c'), 'rev5'))
 
     def corrected_parents(self):
-        return [
+        return (
             # rev3's accessible parent inventories all have rev1a as the last
             # modifier.
-            (['rev1a'], 'rev3'),
+            (('rev1a',), 'rev3'),
             # rev1a features in both rev4's parents but should only appear once
             # in the result
-            (['rev1a'], 'rev4'),
+            (('rev1a',), 'rev4'),
             # rev2c is the head of rev1a and rev2c, the inventory provided
             # per-file last-modified revisions.
-            (['rev2c'], 'rev5')]
+            (('rev2c',), 'rev5'))
 
     def check_regexes(self):
         return [
             "3 inconsistent parents",
-            r"a-file-id version rev3 has parents \['rev2'\] "
-            r"but should have \['rev1a'\]",
-            r"a-file-id version rev5 has parents \['rev2', 'rev2c'\] "
-            r"but should have \['rev2c'\]",
-            r"a-file-id version rev4 has parents \['rev2'\] "
-            r"but should have \['rev1a'\]",
+            r"a-file-id version rev3 has parents \('rev2',\) "
+            r"but should have \('rev1a',\)",
+            r"a-file-id version rev5 has parents \('rev2', 'rev2c'\) "
+            r"but should have \('rev2c',\)",
+            r"a-file-id version rev4 has parents \('rev2',\) "
+            r"but should have \('rev1a',\)",
             ]
 
     def populate_repository(self, repo):
@@ -369,39 +369,39 @@
     """
 
     def all_versions(self):
-        return ['bad-parent', 'good-parent', 'broken-revision']
+        return ('bad-parent', 'good-parent', 'broken-revision')
 
     def populated_parents(self):
-        return [
-            ([], 'bad-parent'),
-            (['bad-parent'], 'good-parent'),
-            (['good-parent', 'bad-parent'], 'broken-revision')]
+        return (
+            ((), 'bad-parent'),
+            (('bad-parent',), 'good-parent'),
+            (('good-parent', 'bad-parent'), 'broken-revision'))
 
     def corrected_parents(self):
-        return [
-            ([], 'bad-parent'),
-            (['bad-parent'], 'good-parent'),
-            (['good-parent'], 'broken-revision')]
+        return (
+            ((), 'bad-parent'),
+            (('bad-parent',), 'good-parent'),
+            (('good-parent',), 'broken-revision'))
 
     def check_regexes(self):
-        return [
+        return (
             '     1 inconsistent parents',
             (r"      \* a-file-id version broken-revision has parents "
-             r"\['good-parent', 'bad-parent'\] but "
-             r"should have \['good-parent'\]")]
+             r"\('good-parent', 'bad-parent'\) but "
+             r"should have \('good-parent',\)"))
 
     def populate_repository(self, repo):
         inv = self.make_one_file_inventory(
-            repo, 'bad-parent', [], root_revision='bad-parent')
-        self.add_revision(repo, 'bad-parent', inv, [])
-        
-        inv = self.make_one_file_inventory(
-            repo, 'good-parent', ['bad-parent'])
-        self.add_revision(repo, 'good-parent', inv, ['bad-parent'])
-        
-        inv = self.make_one_file_inventory(
-            repo, 'broken-revision', ['good-parent', 'bad-parent'])
-        self.add_revision(repo, 'broken-revision', inv, ['good-parent'])
+            repo, 'bad-parent', (), root_revision='bad-parent')
+        self.add_revision(repo, 'bad-parent', inv, ())
+        
+        inv = self.make_one_file_inventory(
+            repo, 'good-parent', ('bad-parent',))
+        self.add_revision(repo, 'good-parent', inv, ('bad-parent',))
+        
+        inv = self.make_one_file_inventory(
+            repo, 'broken-revision', ('good-parent', 'bad-parent'))
+        self.add_revision(repo, 'broken-revision', inv, ('good-parent',))
 
 
 class ClaimedFileParentDidNotModifyFileScenario(BrokenRepoScenario):
@@ -415,43 +415,43 @@
     """
 
     def all_versions(self):
-        return ['basis', 'modified-something-else', 'current']
+        return ('basis', 'modified-something-else', 'current')
 
     def populated_parents(self):
-        return [
-            ([], 'basis'),
-            (['basis'], 'modified-something-else'),
-            (['modified-something-else'], 'current')]
+        return (
+            ((), 'basis'),
+            (('basis',), 'modified-something-else'),
+            (('modified-something-else',), 'current'))
 
     def corrected_parents(self):
-        return [
-            ([], 'basis'),
-            (['basis'], 'modified-something-else'),
-            (['basis'], 'current')]
+        return (
+            ((), 'basis'),
+            (('basis',), 'modified-something-else'),
+            (('basis',), 'current'))
 
     def check_regexes(self):
-        return [
+        return (
             '1 inconsistent parents',
             r"\* a-file-id version current has parents "
-            r"\['modified-something-else'\] but should have \['basis'\]"]
+            r"\('modified-something-else',\) but should have \('basis',\)")
 
     def populate_repository(self, repo):
-        inv = self.make_one_file_inventory(repo, 'basis', [])
-        self.add_revision(repo, 'basis', inv, [])
+        inv = self.make_one_file_inventory(repo, 'basis', ())
+        self.add_revision(repo, 'basis', inv, ())
 
         # 'modified-something-else' is a correctly recorded revision, but it
         # does not modify the file we are looking at, so the inventory for that
         # file in this revision points to 'basis'.
         inv = self.make_one_file_inventory(
-            repo, 'modified-something-else', ['basis'], inv_revision='basis')
-        self.add_revision(repo, 'modified-something-else', inv, ['basis'])
+            repo, 'modified-something-else', ('basis',), inv_revision='basis')
+        self.add_revision(repo, 'modified-something-else', inv, ('basis',))
 
         # The 'current' revision has 'modified-something-else' as its parent,
         # but the 'current' version of 'a-file' should have 'basis' as its
         # parent.
         inv = self.make_one_file_inventory(
-            repo, 'current', ['modified-something-else'])
-        self.add_revision(repo, 'current', inv, ['modified-something-else'])
+            repo, 'current', ('modified-something-else',))
+        self.add_revision(repo, 'current', inv, ('modified-something-else',))
 
 
 class IncorrectlyOrderedParentsScenario(BrokenRepoScenario):
@@ -471,28 +471,28 @@
                 'broken-revision-2-1']
 
     def populated_parents(self):
-        return [
-            ([], 'parent-1'),
-            ([], 'parent-2'),
-            (['parent-2', 'parent-1'], 'broken-revision-1-2'),
-            (['parent-1', 'parent-2'], 'broken-revision-2-1')]
+        return (
+            ((), 'parent-1'),
+            ((), 'parent-2'),
+            (('parent-2', 'parent-1'), 'broken-revision-1-2'),
+            (('parent-1', 'parent-2'), 'broken-revision-2-1'))
 
     def corrected_parents(self):
-        return [
-            ([], 'parent-1'),
-            ([], 'parent-2'),
-            (['parent-1', 'parent-2'], 'broken-revision-1-2'),
-            (['parent-2', 'parent-1'], 'broken-revision-2-1')]
+        return (
+            ((), 'parent-1'),
+            ((), 'parent-2'),
+            (('parent-1', 'parent-2'), 'broken-revision-1-2'),
+            (('parent-2', 'parent-1'), 'broken-revision-2-1'))
 
     def check_regexes(self):
-        return [
+        return (
             "2 inconsistent parents",
             r"\* a-file-id version broken-revision-1-2 has parents "
-            r"\['parent-2', 'parent-1'\] but should have "
-            r"\['parent-1', 'parent-2'\]",
+            r"\('parent-2', 'parent-1'\) but should have "
+            r"\('parent-1', 'parent-2'\)",
             r"\* a-file-id version broken-revision-2-1 has parents "
-            r"\['parent-1', 'parent-2'\] but should have "
-            r"\['parent-2', 'parent-1'\]"]
+            r"\('parent-1', 'parent-2'\) but should have "
+            r"\('parent-2', 'parent-1'\)")
 
     def populate_repository(self, repo):
         inv = self.make_one_file_inventory(repo, 'parent-1', [])

=== modified file 'bzrlib/tests/repository_implementations/helpers.py'
--- bzrlib/tests/repository_implementations/helpers.py	2007-09-24 03:45:21 +0000
+++ bzrlib/tests/repository_implementations/helpers.py	2007-10-17 20:48:21 +0000
@@ -48,13 +48,23 @@
                 "%s isn't a knit format" % self.repository_format)
 
         repo = self.make_repository('broken')
-        inv = inventory.Inventory(revision_id='revision-id')
-        inv.root.revision = 'revision-id'
-        repo.add_inventory('revision-id', inv, [])
-        revision = _mod_revision.Revision('revision-id',
-            committer='jrandom@example.com', timestamp=0,
-            inventory_sha1='', timezone=0, message='message', parent_ids=[])
-        repo.add_revision('revision-id',revision, inv)
+        repo.lock_write()
+        repo.start_write_group()
+        try:
+            inv = inventory.Inventory(revision_id='revision-id')
+            inv.root.revision = 'revision-id'
+            repo.add_inventory('revision-id', inv, [])
+            revision = _mod_revision.Revision('revision-id',
+                committer='jrandom@example.com', timestamp=0,
+                inventory_sha1='', timezone=0, message='message', parent_ids=[])
+            repo.add_revision('revision-id',revision, inv)
+        except:
+            repo.abort_write_group()
+            repo.unlock()
+            raise
+        else:
+            repo.commit_write_group()
+            repo.unlock()
 
         # Change the knit index's record of the parents for 'revision-id' to
         # claim it has a parent, 'incorrect-parent', that doesn't exist in this

=== modified file 'bzrlib/tests/repository_implementations/test_check.py'
--- bzrlib/tests/repository_implementations/test_check.py	2007-10-05 01:56:16 +0000
+++ bzrlib/tests/repository_implementations/test_check.py	2007-10-17 09:39:41 +0000
@@ -39,7 +39,7 @@
         """
         repo = self.make_repo_with_extra_ghost_index()
         self.assertEqual(
-            [('revision-id', ['incorrect-parent'], [])],
+            [('revision-id', ('incorrect-parent',), ())],
             list(repo._find_inconsistent_revision_parents()))
 
     def test__check_for_inconsistent_revision_parents(self):
@@ -59,7 +59,11 @@
         if not repo.revision_graph_can_have_wrong_parents():
             raise TestNotApplicable(
                 '%r cannot have corrupt revision index.' % repo)
-        repo._check_for_inconsistent_revision_parents()  # nothing happens
+        repo.lock_read()
+        try:
+            repo._check_for_inconsistent_revision_parents()  # nothing happens
+        finally:
+            repo.unlock()
 
     def test_check_reports_bad_ancestor(self):
         repo = self.make_repo_with_extra_ghost_index()
@@ -75,5 +79,5 @@
         self.assertContainsRe(
             log,
             "revision-id has wrong parents in index: "
-            r"\['incorrect-parent'\] should be \[\]")
+            r"\('incorrect-parent',\) should be \(\)")
 

=== modified file 'bzrlib/tests/repository_implementations/test_check_reconcile.py'
--- bzrlib/tests/repository_implementations/test_check_reconcile.py	2007-10-03 10:54:07 +0000
+++ bzrlib/tests/repository_implementations/test_check_reconcile.py	2007-10-17 09:39:41 +0000
@@ -103,8 +103,8 @@
                     "Format does not support text parent reconciliation")
 
     def file_parents(self, repo, revision_id):
-        return repo.weave_store.get_weave('a-file-id',
-            repo.get_transaction()).get_parents(revision_id)
+        return tuple(repo.weave_store.get_weave('a-file-id',
+            repo.get_transaction()).get_parents(revision_id))
 
     def assertParentsMatch(self, expected_parents_for_versions, repo,
                            when_description):

=== modified file 'bzrlib/tests/repository_implementations/test_commit_builder.py'
--- bzrlib/tests/repository_implementations/test_commit_builder.py	2007-10-19 17:00:10 +0000
+++ bzrlib/tests/repository_implementations/test_commit_builder.py	2007-10-22 01:23:51 +0000
@@ -213,11 +213,15 @@
         self.assertEqual(rev_id, basis_tree.inventory.root.revision)
 
     def _get_revtrees(self, tree, revision_ids):
-        trees = list(tree.branch.repository.revision_trees(revision_ids))
-        for tree in trees:
-            tree.lock_read()
-            self.addCleanup(tree.unlock)
-        return trees
+        tree.lock_read()
+        try:
+            trees = list(tree.branch.repository.revision_trees(revision_ids))
+            for _tree in trees:
+                _tree.lock_read()
+                self.addCleanup(_tree.unlock)
+            return trees
+        finally:
+            tree.unlock()
 
     def test_last_modified_revision_after_commit_root_unchanged(self):
         # commiting without changing the root does not change the 

=== modified file 'bzrlib/tests/repository_implementations/test_fileid_involved.py'
--- bzrlib/tests/repository_implementations/test_fileid_involved.py	2007-08-03 01:07:18 +0000
+++ bzrlib/tests/repository_implementations/test_fileid_involved.py	2007-10-17 09:39:41 +0000
@@ -139,7 +139,8 @@
     def test_fileids_altered_between_two_revs(self):
         def foo(old, new):
             print set(self.branch.repository.get_ancestry(new)).difference(set(self.branch.repository.get_ancestry(old)))
-
+        self.branch.lock_read()
+        self.addCleanup(self.branch.unlock)
         self.assertEqual(
             {'b-file-id-2006-01-01-defg':set(['rev-J']),
              'c-funky<file-id>quiji%bo':set(['rev-K'])
@@ -180,11 +181,13 @@
         return result
 
     def test_fileids_altered_by_revision_ids(self):
+        self.branch.lock_read()
+        self.addCleanup(self.branch.unlock)
         self.assertEqual(
             {'a-file-id-2006-01-01-abcd':set(['rev-A']),
              'b-file-id-2006-01-01-defg': set(['rev-A']),
              'c-funky<file-id>quiji%bo': set(['rev-A']),
-             }, 
+             },
             self.fileids_altered_by_revision_ids(["rev-A"]))
         self.assertEqual(
             {'a-file-id-2006-01-01-abcd':set(['rev-B'])
@@ -201,6 +204,8 @@
         # comparing the trees - no less, and no more. This is correct 
         # because in our sample data we do not revert any file ids along
         # the revision history.
+        self.branch.lock_read()
+        self.addCleanup(self.branch.unlock)
         pp=[]
         history = self.branch.revision_history( )
 
@@ -237,6 +242,8 @@
                               % self.repository_format)
 
         repo = main_wt.branch.repository
+        repo.lock_read()
+        self.addCleanup(repo.unlock)
         file_ids = repo.fileids_altered_by_revision_ids([revision_id])
         root_id = main_wt.basis_tree().path2id('')
         if root_id in file_ids:
@@ -290,6 +297,8 @@
         # this tests that fileids_alteted_by_revision_ids returns 
         # more information than compare_tree can, because it 
         # sees each change rather than the aggregate delta.
+        self.branch.lock_read()
+        self.addCleanup(self.branch.unlock)
         history = self.branch.revision_history()
         old_rev = history[0]
         new_rev = history[1]

=== modified file 'bzrlib/tests/repository_implementations/test_repository.py'
--- bzrlib/tests/repository_implementations/test_repository.py	2007-10-12 08:18:54 +0000
+++ bzrlib/tests/repository_implementations/test_repository.py	2007-10-17 09:39:41 +0000
@@ -61,6 +61,8 @@
         bzrdirb = self.make_bzrdir('b')
         repo_b = tree_a.branch.repository.clone(bzrdirb)
         tree_b = repo_b.revision_tree('rev1')
+        tree_b.lock_read()
+        self.addCleanup(tree_b.unlock)
         tree_b.get_file_text('file1')
         rev1 = repo_b.get_revision('rev1')
 
@@ -421,7 +423,20 @@
             self.assertContainsRe(str(e), 'get_data_stream')
             raise TestSkipped('This format does not support streaming.')
 
-        dest_repo.insert_data_stream(stream)
+        dest_repo.lock_write()
+        try:
+            dest_repo.start_write_group()
+            try:
+                dest_repo.insert_data_stream(stream)
+            except:
+                dest_repo.abort_write_group()
+                raise
+            else:
+                dest_repo.commit_write_group()
+        finally:
+            dest_repo.unlock()
+        # reopen to be sure it was added.
+        dest_repo = dest_repo.bzrdir.open_repository()
         self.assertTrue(dest_repo.has_revision('rev_id'))
 
     def test_get_serializer_format(self):
@@ -438,6 +453,8 @@
         self.build_tree_contents([('tree/file1', 'baz')])
         tree.commit('rev2', rev_id='rev2')
         repository = tree.branch.repository
+        repository.lock_read()
+        self.addCleanup(repository.unlock)
         extracted = dict((i, ''.join(b)) for i, b in
                          repository.iter_files_bytes(
                          [('file1-id', 'rev1', 'file1-old'),
@@ -494,6 +511,8 @@
     def test_get_graph(self):
         """Bare-bones smoketest that all repositories implement get_graph."""
         repo = self.make_repository('repo')
+        repo.lock_read()
+        self.addCleanup(repo.unlock)
         repo.get_graph()
 
     def test_implements_revision_graph_can_have_wrong_parents(self):
@@ -503,14 +522,16 @@
         """
         repo = self.make_repository('.')
         # This should work, not raise NotImplementedError:
-        result = repo.revision_graph_can_have_wrong_parents()
-        if result:
-            # If true, then this repo must also implement
-            # _find_inconsistent_revision_parents and
-            # _check_for_inconsistent_revision_parents.  So calling these
-            # should also not raise NotImplementedError.
-            list(repo._find_inconsistent_revision_parents())
-            repo._check_for_inconsistent_revision_parents()
+        if not repo.revision_graph_can_have_wrong_parents():
+            return
+        repo.lock_read()
+        self.addCleanup(repo.unlock)
+        # This repo must also implement
+        # _find_inconsistent_revision_parents and
+        # _check_for_inconsistent_revision_parents.  So calling these
+        # should not raise NotImplementedError.
+        list(repo._find_inconsistent_revision_parents())
+        repo._check_for_inconsistent_revision_parents()
 
 
 class TestRepositoryLocking(TestCaseWithRepository):
@@ -606,6 +627,8 @@
     def test_revision_trees(self):
         revision_ids = ['rev1', 'rev2', 'rev3', 'rev4']
         repository = self.bzrdir.open_repository()
+        repository.lock_read()
+        self.addCleanup(repository.unlock)
         trees1 = list(repository.revision_trees(revision_ids))
         trees2 = [repository.revision_tree(t) for t in revision_ids]
         assert len(trees1) == len(trees2)
@@ -614,6 +637,8 @@
 
     def test_get_deltas_for_revisions(self):
         repository = self.bzrdir.open_repository()
+        repository.lock_read()
+        self.addCleanup(repository.unlock)
         revisions = [repository.get_revision(r) for r in 
                      ['rev1', 'rev2', 'rev3', 'rev4']]
         deltas1 = list(repository.get_deltas_for_revisions(revisions))

=== modified file 'bzrlib/tests/test_repository.py'
--- bzrlib/tests/test_repository.py	2007-10-19 17:14:33 +0000
+++ bzrlib/tests/test_repository.py	2007-10-24 06:48:13 +0000
@@ -22,6 +22,7 @@
 also see this file.
 """
 
+import md5
 from stat import S_ISDIR
 from StringIO import StringIO
 
@@ -31,6 +32,7 @@
                            UnknownFormatError,
                            UnsupportedFormatError,
                            )
+from bzrlib.index import GraphIndex, InMemoryGraphIndex
 from bzrlib.repository import RepositoryFormat
 from bzrlib.smart import server
 from bzrlib.tests import (
@@ -51,7 +53,7 @@
     upgrade,
     workingtree,
     )
-from bzrlib.repofmt import knitrepo, weaverepo
+from bzrlib.repofmt import knitrepo, weaverepo, pack_repo
 
 
 class TestDefaultFormat(TestCase):
@@ -670,6 +672,7 @@
 
 
 class TestWithBrokenRepo(TestCaseWithTransport):
+    """These tests seem to be more appropriate as interface tests?"""
 
     def make_broken_repository(self):
         # XXX: This function is borrowed from Aaron's "Reconcile can fix bad
@@ -749,6 +752,585 @@
         broken_repo = self.make_broken_repository()
         empty_repo = self.make_repository('empty-repo')
         stream = broken_repo.get_data_stream(['rev1a', 'rev2', 'rev3'])
-        self.assertRaises(
-            errors.KnitCorrupt, empty_repo.insert_data_stream, stream)
-
+        empty_repo.lock_write()
+        self.addCleanup(empty_repo.unlock)
+        empty_repo.start_write_group()
+        try:
+            self.assertRaises(
+                errors.KnitCorrupt, empty_repo.insert_data_stream, stream)
+        finally:
+            empty_repo.abort_write_group()
+
+
+class TestExperimentalNoSubtrees(TestCaseWithTransport):
+
+    def get_format(self):
+        return bzrdir.format_registry.make_bzrdir('experimental')
+
+    def test_disk_layout(self):
+        format = self.get_format()
+        repo = self.make_repository('.', format=format)
+        # in case of side effects of locking.
+        repo.lock_write()
+        repo.unlock()
+        t = repo.bzrdir.get_repository_transport(None)
+        self.check_format(t)
+        # XXX: no locks left when unlocked at the moment
+        # self.assertEqualDiff('', t.get('lock').read())
+        self.check_databases(t)
+
+    def check_format(self, t):
+        self.assertEqualDiff('Bazaar Experimental no-subtrees\n',
+                             t.get('format').read())
+
+    def assertHasKndx(self, t, knit_name):
+        """Assert that knit_name exists on t."""
+        self.assertEqualDiff('# bzr knit index 8\n',
+                             t.get(knit_name + '.kndx').read())
+
+    def assertHasNoKndx(self, t, knit_name):
+        """Assert that knit_name has no index on t."""
+        self.assertFalse(t.has(knit_name + '.kndx'))
+
+    def assertHasNoKnit(self, t, knit_name):
+        """Assert that knit_name exists on t."""
+        # no default content
+        self.assertFalse(t.has(knit_name + '.knit'))
+
+    def check_databases(self, t):
+        """check knit content for a repository."""
+        # check conversion worked
+        self.assertHasNoKndx(t, 'inventory')
+        self.assertHasNoKnit(t, 'inventory')
+        self.assertHasNoKndx(t, 'revisions')
+        self.assertHasNoKnit(t, 'revisions')
+        self.assertHasNoKndx(t, 'signatures')
+        self.assertHasNoKnit(t, 'signatures')
+        self.assertFalse(t.has('knits'))
+        # revision-indexes file-container directory
+        self.assertEqual([],
+            list(GraphIndex(t, 'pack-names', None).iter_all_entries()))
+        self.assertTrue(S_ISDIR(t.stat('packs').st_mode))
+        self.assertTrue(S_ISDIR(t.stat('upload').st_mode))
+        self.assertTrue(S_ISDIR(t.stat('indices').st_mode))
+        self.assertTrue(S_ISDIR(t.stat('obsolete_packs').st_mode))
+
+    def test_shared_disk_layout(self):
+        format = self.get_format()
+        repo = self.make_repository('.', shared=True, format=format)
+        # we want:
+        t = repo.bzrdir.get_repository_transport(None)
+        self.check_format(t)
+        # XXX: no locks left when unlocked at the moment
+        # self.assertEqualDiff('', t.get('lock').read())
+        # We should have a 'shared-storage' marker file.
+        self.assertEqualDiff('', t.get('shared-storage').read())
+        self.check_databases(t)
+
+    def test_shared_no_tree_disk_layout(self):
+        format = self.get_format()
+        repo = self.make_repository('.', shared=True, format=format)
+        repo.set_make_working_trees(False)
+        # we want:
+        t = repo.bzrdir.get_repository_transport(None)
+        self.check_format(t)
+        # XXX: no locks left when unlocked at the moment
+        # self.assertEqualDiff('', t.get('lock').read())
+        # We should have a 'shared-storage' marker file.
+        self.assertEqualDiff('', t.get('shared-storage').read())
+        # We should have a marker for the no-working-trees flag.
+        self.assertEqualDiff('', t.get('no-working-trees').read())
+        # The marker should go when we toggle the setting.
+        repo.set_make_working_trees(True)
+        self.assertFalse(t.has('no-working-trees'))
+        self.check_databases(t)
+
+    def test_adding_revision_creates_pack_indices(self):
+        format = self.get_format()
+        tree = self.make_branch_and_tree('.', format=format)
+        trans = tree.branch.repository.bzrdir.get_repository_transport(None)
+        self.assertEqual([],
+            list(GraphIndex(trans, 'pack-names', None).iter_all_entries()))
+        tree.commit('foobarbaz')
+        index = GraphIndex(trans, 'pack-names', None)
+        index_nodes = list(index.iter_all_entries())
+        self.assertEqual(1, len(index_nodes))
+        node = index_nodes[0]
+        name = node[1][0]
+        # the pack sizes should be listed in the index
+        pack_value = node[2]
+        sizes = [int(digits) for digits in pack_value.split(' ')]
+        for size, suffix in zip(sizes, ['.rix', '.iix', '.tix', '.six']):
+            stat = trans.stat('indices/%s%s' % (name, suffix))
+            self.assertEqual(size, stat.st_size)
+
+    def test_pulling_nothing_leads_to_no_new_names(self):
+        format = self.get_format()
+        tree1 = self.make_branch_and_tree('1', format=format)
+        tree2 = self.make_branch_and_tree('2', format=format)
+        tree1.branch.repository.fetch(tree2.branch.repository)
+        trans = tree1.branch.repository.bzrdir.get_repository_transport(None)
+        self.assertEqual([],
+            list(GraphIndex(trans, 'pack-names', None).iter_all_entries()))
+
+    def test_commit_across_pack_shape_boundary_autopacks(self):
+        format = self.get_format()
+        tree = self.make_branch_and_tree('.', format=format)
+        trans = tree.branch.repository.bzrdir.get_repository_transport(None)
+        # This test could be a little cheaper by replacing the packs
+        # attribute on the repository to allow a different pack distribution
+        # and max packs policy - so we are checking the policy is honoured
+        # in the test. But for now 11 commits is not a big deal in a single
+        # test.
+        for x in range(9):
+            tree.commit('commit %s' % x)
+        # there should be 9 packs:
+        index = GraphIndex(trans, 'pack-names', None)
+        self.assertEqual(9, len(list(index.iter_all_entries())))
+        # committing one more should coalesce to 1 of 10.
+        tree.commit('commit triggering pack')
+        index = GraphIndex(trans, 'pack-names', None)
+        self.assertEqual(1, len(list(index.iter_all_entries())))
+        # packing should not damage data
+        tree = tree.bzrdir.open_workingtree()
+        check_result = tree.branch.repository.check(
+            [tree.branch.last_revision()])
+        # XXX: Todo check packs obsoleted correctly - old packs and indices
+        # in the obsolete_packs directory.
+        large_pack_name = list(index.iter_all_entries())[0][1][0]
+        # finally, committing again should not touch the large pack.
+        tree.commit('commit not triggering pack')
+        index = GraphIndex(trans, 'pack-names', None)
+        self.assertEqual(2, len(list(index.iter_all_entries())))
+        pack_names = [node[1][0] for node in index.iter_all_entries()]
+        self.assertTrue(large_pack_name in pack_names)
+
+    def test_pack_after_two_commits_packs_everything(self):
+        format = self.get_format()
+        tree = self.make_branch_and_tree('.', format=format)
+        trans = tree.branch.repository.bzrdir.get_repository_transport(None)
+        tree.commit('start')
+        tree.commit('more work')
+        tree.branch.repository.pack()
+        # there should be 1 pack:
+        index = GraphIndex(trans, 'pack-names', None)
+        self.assertEqual(1, len(list(index.iter_all_entries())))
+        self.assertEqual(2, len(tree.branch.repository.all_revision_ids()))
+
+    def test_pack_repositories_support_multiple_write_locks(self):
+        format = self.get_format()
+        self.make_repository('.', shared=True, format=format)
+        r1 = repository.Repository.open('.')
+        r2 = repository.Repository.open('.')
+        r1.lock_write()
+        self.addCleanup(r1.unlock)
+        r2.lock_write()
+        r2.unlock()
+
+    def _add_text(self, repo, fileid):
+        """Add a text to the repository within a write group."""
+        vf =repo.weave_store.get_weave(fileid, repo.get_transaction())
+        vf.add_lines('samplerev+' + fileid, [], [])
+
+    def test_concurrent_writers_merge_new_packs(self):
+        format = self.get_format()
+        self.make_repository('.', shared=True, format=format)
+        r1 = repository.Repository.open('.')
+        r2 = repository.Repository.open('.')
+        r1.lock_write()
+        try:
+            # access enough data to load the names list
+            list(r1.all_revision_ids())
+            r2.lock_write()
+            try:
+                # access enough data to load the names list
+                list(r2.all_revision_ids())
+                r1.start_write_group()
+                try:
+                    r2.start_write_group()
+                    try:
+                        self._add_text(r1, 'fileidr1')
+                        self._add_text(r2, 'fileidr2')
+                    except:
+                        r2.abort_write_group()
+                        raise
+                except:
+                    r1.abort_write_group()
+                    raise
+                # both r1 and r2 have open write groups with data in them
+                # created while the other's write group was open.
+                # Commit both which requires a merge to the pack-names.
+                try:
+                    r1.commit_write_group()
+                except:
+                    r1.abort_write_group()
+                    r2.abort_write_group()
+                    raise
+                r2.commit_write_group()
+                # tell r1 to reload from disk
+                r1._pack_collection.reset()
+                # Now both repositories should know about both names
+                r1._pack_collection.ensure_loaded()
+                r2._pack_collection.ensure_loaded()
+                self.assertEqual(r1._pack_collection.names(), r2._pack_collection.names())
+                self.assertEqual(2, len(r1._pack_collection.names()))
+            finally:
+                r1.unlock()
+        finally:
+            r2.unlock()
+
+    def test_concurrent_writer_second_preserves_dropping_a_pack(self):
+        format = self.get_format()
+        self.make_repository('.', shared=True, format=format)
+        r1 = repository.Repository.open('.')
+        r2 = repository.Repository.open('.')
+        # add a pack to drop
+        r1.lock_write()
+        try:
+            r1.start_write_group()
+            try:
+                self._add_text(r1, 'fileidr1')
+            except:
+                r1.abort_write_group()
+                raise
+            else:
+                r1.commit_write_group()
+            r1._pack_collection.ensure_loaded()
+            name_to_drop = r1._pack_collection.all_packs()[0].name
+        finally:
+            r1.unlock()
+        r1.lock_write()
+        try:
+            # access enough data to load the names list
+            list(r1.all_revision_ids())
+            r2.lock_write()
+            try:
+                # access enough data to load the names list
+                list(r2.all_revision_ids())
+                r1._pack_collection.ensure_loaded()
+                try:
+                    r2.start_write_group()
+                    try:
+                        # in r1, drop the pack
+                        r1._pack_collection._remove_pack_from_memory(
+                            r1._pack_collection.get_pack_by_name(name_to_drop))
+                        # in r2, add a pack
+                        self._add_text(r2, 'fileidr2')
+                    except:
+                        r2.abort_write_group()
+                        raise
+                except:
+                    r1._pack_collection.reset()
+                    raise
+                # r1 has a changed names list, and r2 an open write groups with
+                # changes.
+                # save r1, and then commit the r2 write group, which requires a
+                # merge to the pack-names, which should not reinstate
+                # name_to_drop
+                try:
+                    r1._pack_collection._save_pack_names()
+                    r1._pack_collection.reset()
+                except:
+                    r2.abort_write_group()
+                    raise
+                try:
+                    r2.commit_write_group()
+                except:
+                    r2.abort_write_group()
+                    raise
+                # Now both repositories should now about just one name.
+                r1._pack_collection.ensure_loaded()
+                r2._pack_collection.ensure_loaded()
+                self.assertEqual(r1._pack_collection.names(), r2._pack_collection.names())
+                self.assertEqual(1, len(r1._pack_collection.names()))
+                self.assertFalse(name_to_drop in r1._pack_collection.names())
+            finally:
+                r1.unlock()
+        finally:
+            r2.unlock()
+
+    def test_lock_write_does_not_physically_lock(self):
+        repo = self.make_repository('.', format=self.get_format())
+        repo.lock_write()
+        self.addCleanup(repo.unlock)
+        self.assertFalse(repo.get_physical_lock_status())
+
+    def prepare_for_break_lock(self):
+        # Setup the global ui factory state so that a break-lock method call
+        # will find usable input in the input stream.
+        old_factory = bzrlib.ui.ui_factory
+        def restoreFactory():
+            bzrlib.ui.ui_factory = old_factory
+        self.addCleanup(restoreFactory)
+        bzrlib.ui.ui_factory = bzrlib.ui.SilentUIFactory()
+        bzrlib.ui.ui_factory.stdin = StringIO("y\n")
+
+    def test_break_lock_breaks_physical_lock(self):
+        repo = self.make_repository('.', format=self.get_format())
+        repo._pack_collection.lock_names()
+        repo2 = repository.Repository.open('.')
+        self.assertTrue(repo.get_physical_lock_status())
+        self.prepare_for_break_lock()
+        repo2.break_lock()
+        self.assertFalse(repo.get_physical_lock_status())
+
+    def test_broken_physical_locks_error_on__unlock_names_lock(self):
+        repo = self.make_repository('.', format=self.get_format())
+        repo._pack_collection.lock_names()
+        self.assertTrue(repo.get_physical_lock_status())
+        repo2 = repository.Repository.open('.')
+        self.prepare_for_break_lock()
+        repo2.break_lock()
+        self.assertRaises(errors.LockBroken, repo._pack_collection._unlock_names)
+
+
+class TestExperimentalSubtrees(TestExperimentalNoSubtrees):
+
+    def get_format(self):
+        return bzrdir.format_registry.make_bzrdir('experimental-subtree')
+
+    def check_format(self, t):
+        self.assertEqualDiff('Bazaar Experimental subtrees\n',
+                             t.get('format').read())
+
+
+class TestRepositoryPackCollection(TestCaseWithTransport):
+
+    def get_format(self):
+        return bzrdir.format_registry.make_bzrdir('experimental')
+
+    def test__max_pack_count(self):
+        """The maximum pack count is a function of the number of revisions."""
+        format = self.get_format()
+        repo = self.make_repository('.', format=format)
+        packs = repo._pack_collection
+        # no revisions - one pack, so that we can have a revision free repo
+        # without it blowing up
+        self.assertEqual(1, packs._max_pack_count(0))
+        # after that the sum of the digits, - check the first 1-9
+        self.assertEqual(1, packs._max_pack_count(1))
+        self.assertEqual(2, packs._max_pack_count(2))
+        self.assertEqual(3, packs._max_pack_count(3))
+        self.assertEqual(4, packs._max_pack_count(4))
+        self.assertEqual(5, packs._max_pack_count(5))
+        self.assertEqual(6, packs._max_pack_count(6))
+        self.assertEqual(7, packs._max_pack_count(7))
+        self.assertEqual(8, packs._max_pack_count(8))
+        self.assertEqual(9, packs._max_pack_count(9))
+        # check the boundary cases with two digits for the next decade
+        self.assertEqual(1, packs._max_pack_count(10))
+        self.assertEqual(2, packs._max_pack_count(11))
+        self.assertEqual(10, packs._max_pack_count(19))
+        self.assertEqual(2, packs._max_pack_count(20))
+        self.assertEqual(3, packs._max_pack_count(21))
+        # check some arbitrary big numbers
+        self.assertEqual(25, packs._max_pack_count(112894))
+
+    def test_pack_distribution_zero(self):
+        format = self.get_format()
+        repo = self.make_repository('.', format=format)
+        packs = repo._pack_collection
+        self.assertEqual([0], packs.pack_distribution(0))
+        
+    def test_pack_distribution_one_to_nine(self):
+        format = self.get_format()
+        repo = self.make_repository('.', format=format)
+        packs = repo._pack_collection
+        self.assertEqual([1],
+            packs.pack_distribution(1))
+        self.assertEqual([1, 1],
+            packs.pack_distribution(2))
+        self.assertEqual([1, 1, 1],
+            packs.pack_distribution(3))
+        self.assertEqual([1, 1, 1, 1],
+            packs.pack_distribution(4))
+        self.assertEqual([1, 1, 1, 1, 1],
+            packs.pack_distribution(5))
+        self.assertEqual([1, 1, 1, 1, 1, 1],
+            packs.pack_distribution(6))
+        self.assertEqual([1, 1, 1, 1, 1, 1, 1],
+            packs.pack_distribution(7))
+        self.assertEqual([1, 1, 1, 1, 1, 1, 1, 1],
+            packs.pack_distribution(8))
+        self.assertEqual([1, 1, 1, 1, 1, 1, 1, 1, 1],
+            packs.pack_distribution(9))
+
+    def test_pack_distribution_stable_at_boundaries(self):
+        """When there are multi-rev packs the counts are stable."""
+        format = self.get_format()
+        repo = self.make_repository('.', format=format)
+        packs = repo._pack_collection
+        # in 10s:
+        self.assertEqual([10], packs.pack_distribution(10))
+        self.assertEqual([10, 1], packs.pack_distribution(11))
+        self.assertEqual([10, 10], packs.pack_distribution(20))
+        self.assertEqual([10, 10, 1], packs.pack_distribution(21))
+        # 100s
+        self.assertEqual([100], packs.pack_distribution(100))
+        self.assertEqual([100, 1], packs.pack_distribution(101))
+        self.assertEqual([100, 10, 1], packs.pack_distribution(111))
+        self.assertEqual([100, 100], packs.pack_distribution(200))
+        self.assertEqual([100, 100, 1], packs.pack_distribution(201))
+        self.assertEqual([100, 100, 10, 1], packs.pack_distribution(211))
+
+    def test_plan_pack_operations_2009_revisions_skip_all_packs(self):
+        format = self.get_format()
+        repo = self.make_repository('.', format=format)
+        packs = repo._pack_collection
+        existing_packs = [(2000, "big"), (9, "medium")]
+        # rev count - 2009 -> 2x1000 + 9x1
+        pack_operations = packs.plan_autopack_combinations(
+            existing_packs, [1000, 1000, 1, 1, 1, 1, 1, 1, 1, 1, 1])
+        self.assertEqual([], pack_operations)
+
+    def test_plan_pack_operations_2010_revisions_skip_all_packs(self):
+        format = self.get_format()
+        repo = self.make_repository('.', format=format)
+        packs = repo._pack_collection
+        existing_packs = [(2000, "big"), (9, "medium"), (1, "single")]
+        # rev count - 2010 -> 2x1000 + 1x10
+        pack_operations = packs.plan_autopack_combinations(
+            existing_packs, [1000, 1000, 10])
+        self.assertEqual([], pack_operations)
+
+    def test_plan_pack_operations_2010_combines_smallest_two(self):
+        format = self.get_format()
+        repo = self.make_repository('.', format=format)
+        packs = repo._pack_collection
+        existing_packs = [(1999, "big"), (9, "medium"), (1, "single2"),
+            (1, "single1")]
+        # rev count - 2010 -> 2x1000 + 1x10 (3)
+        pack_operations = packs.plan_autopack_combinations(
+            existing_packs, [1000, 1000, 10])
+        self.assertEqual([[2, ["single2", "single1"]], [0, []]], pack_operations)
+
+    def test_all_packs_none(self):
+        format = self.get_format()
+        tree = self.make_branch_and_tree('.', format=format)
+        tree.lock_read()
+        self.addCleanup(tree.unlock)
+        packs = tree.branch.repository._pack_collection
+        packs.ensure_loaded()
+        self.assertEqual([], packs.all_packs())
+
+    def test_all_packs_one(self):
+        format = self.get_format()
+        tree = self.make_branch_and_tree('.', format=format)
+        tree.commit('start')
+        tree.lock_read()
+        self.addCleanup(tree.unlock)
+        packs = tree.branch.repository._pack_collection
+        packs.ensure_loaded()
+        self.assertEqual([
+            packs.get_pack_by_name(packs.names()[0])],
+            packs.all_packs())
+
+    def test_all_packs_two(self):
+        format = self.get_format()
+        tree = self.make_branch_and_tree('.', format=format)
+        tree.commit('start')
+        tree.commit('continue')
+        tree.lock_read()
+        self.addCleanup(tree.unlock)
+        packs = tree.branch.repository._pack_collection
+        packs.ensure_loaded()
+        self.assertEqual([
+            packs.get_pack_by_name(packs.names()[0]),
+            packs.get_pack_by_name(packs.names()[1]),
+            ], packs.all_packs())
+
+    def test_get_pack_by_name(self):
+        format = self.get_format()
+        tree = self.make_branch_and_tree('.', format=format)
+        tree.commit('start')
+        tree.lock_read()
+        self.addCleanup(tree.unlock)
+        packs = tree.branch.repository._pack_collection
+        packs.ensure_loaded()
+        name = packs.names()[0]
+        pack_1 = packs.get_pack_by_name(name)
+        # the pack should be correctly initialised
+        rev_index = GraphIndex(packs._index_transport, name + '.rix',
+            packs._names[name][0])
+        inv_index = GraphIndex(packs._index_transport, name + '.iix',
+            packs._names[name][1])
+        txt_index = GraphIndex(packs._index_transport, name + '.tix',
+            packs._names[name][2])
+        sig_index = GraphIndex(packs._index_transport, name + '.six',
+            packs._names[name][3])
+        self.assertEqual(pack_repo.ExistingPack(packs._pack_transport,
+            name, rev_index, inv_index, txt_index, sig_index), pack_1)
+        # and the same instance should be returned on successive calls.
+        self.assertTrue(pack_1 is packs.get_pack_by_name(name))
+
+
+class TestPack(TestCaseWithTransport):
+    """Tests for the Pack object."""
+
+    def assertCurrentlyEqual(self, left, right):
+        self.assertTrue(left == right)
+        self.assertTrue(right == left)
+        self.assertFalse(left != right)
+        self.assertFalse(right != left)
+
+    def assertCurrentlyNotEqual(self, left, right):
+        self.assertFalse(left == right)
+        self.assertFalse(right == left)
+        self.assertTrue(left != right)
+        self.assertTrue(right != left)
+
+    def test___eq____ne__(self):
+        left = pack_repo.ExistingPack('', '', '', '', '', '')
+        right = pack_repo.ExistingPack('', '', '', '', '', '')
+        self.assertCurrentlyEqual(left, right)
+        # change all attributes and ensure equality changes as we do.
+        left.revision_index = 'a'
+        self.assertCurrentlyNotEqual(left, right)
+        right.revision_index = 'a'
+        self.assertCurrentlyEqual(left, right)
+        left.inventory_index = 'a'
+        self.assertCurrentlyNotEqual(left, right)
+        right.inventory_index = 'a'
+        self.assertCurrentlyEqual(left, right)
+        left.text_index = 'a'
+        self.assertCurrentlyNotEqual(left, right)
+        right.text_index = 'a'
+        self.assertCurrentlyEqual(left, right)
+        left.signature_index = 'a'
+        self.assertCurrentlyNotEqual(left, right)
+        right.signature_index = 'a'
+        self.assertCurrentlyEqual(left, right)
+        left.name = 'a'
+        self.assertCurrentlyNotEqual(left, right)
+        right.name = 'a'
+        self.assertCurrentlyEqual(left, right)
+        left.transport = 'a'
+        self.assertCurrentlyNotEqual(left, right)
+        right.transport = 'a'
+        self.assertCurrentlyEqual(left, right)
+
+    def test_file_name(self):
+        pack = pack_repo.ExistingPack('', 'a_name', '', '', '', '')
+        self.assertEqual('a_name.pack', pack.file_name())
+
+
+class TestNewPack(TestCaseWithTransport):
+    """Tests for pack_repo.NewPack."""
+
+    def test_new_instance_attributes(self):
+        upload_transport = self.get_transport('upload')
+        pack_transport = self.get_transport('pack')
+        index_transport = self.get_transport('index')
+        upload_transport.mkdir('.')
+        pack = pack_repo.NewPack(upload_transport, index_transport,
+            pack_transport)
+        self.assertIsInstance(pack.revision_index, InMemoryGraphIndex)
+        self.assertIsInstance(pack.inventory_index, InMemoryGraphIndex)
+        self.assertIsInstance(pack._hash, type(md5.new()))
+        self.assertTrue(pack.upload_transport is upload_transport)
+        self.assertTrue(pack.index_transport is index_transport)
+        self.assertTrue(pack.pack_transport is pack_transport)
+        self.assertEqual(None, pack.index_sizes)
+        self.assertEqual(20, len(pack.random_name))
+        self.assertIsInstance(pack.random_name, str)
+        self.assertIsInstance(pack.start_time, float)

=== modified file 'doc/developers/index.txt'
--- doc/developers/index.txt	2007-08-13 10:20:11 +0000
+++ doc/developers/index.txt	2007-10-24 01:52:12 +0000
@@ -31,5 +31,11 @@
 
 * `Repositories <repository.html>`_ |--| What repositories do and are used for.
 
+Data formats
+============
+
+* `Knit pack repositories <knitpack.html>`_ |--| KnitPack repositories 
+  (new in Bazaar 0.92).
+
 
 .. |--| unicode:: U+2014

=== added file 'doc/developers/knitpack.txt'
--- doc/developers/knitpack.txt	1970-01-01 00:00:00 +0000
+++ doc/developers/knitpack.txt	2007-10-24 02:14:11 +0000
@@ -0,0 +1,143 @@
+KnitPack repository format
+==========================
+
+Bazaar 0.92 adds a new format (experimental at first) implemented in
+``bzrlib.repofmt.pack_repo.py``.  
+
+This format provides a knit-like interface which is quite compatible
+with knit format repositories: you can get a VersionedFile for a
+particular file-id, or for revisions, or for the inventory, even though
+these do not correspond to single files on disk.
+
+The on-disk format is that the repository directory contains these
+files and subdirectories:
+
+==================== =============================================
+packs/               completed readonly packs
+indices/             indices for completed packs
+upload/              temporary files for packs currently being 
+                     written
+obsolete_packs/      packs that have been repacked and are no 
+                     longer normally needed
+pack-names           index of all live packs
+lock/                lockdir
+==================== =============================================
+
+Note that for consistency we always write "indices" not "indexes".
+
+This is implemented on top of pack files, which are written once from
+start to end, then left alone.  A pack consists of a body file, plus
+several index files.  There are four index files for each pack, which
+have the same basename and an extension indicating the purpose of the
+index:
+
+======== ========== ======================== ==========================
+extn     Purpose    Key                      References
+======== ========== ======================== ==========================
+``.tix`` File texts ``file_id, revision_id`` per-file parents,
+                                             compression basis
+                                             per-file parents
+``.six`` Signatures ``revision_id,``         -
+``.rix`` Revisions  ``revision_id,``         revision parents
+``.iix`` Inventory  ``revision_id,``         revision parents,
+                                             compression base
+======== ========== ======================== ==========================
+
+Indices are accessed through the ``bzrlib.index.GraphIndex`` class.  
+Indices are stored as sorted files on disk.  Each line is one record,
+and contains:
+
+ * key fields
+ * a value string - for all these indices, this is an ascii decimal pair
+   of "offset length" giving the position of the refenced data within 
+   the pack body file
+ * a list of zero or more reference lists
+
+The reference lists let a graph be stored within the index.  Each
+reference list entry points to another entry in the same index.  The
+references are represented as a byte offset for the target within the
+index file.
+
+When a compression base is given, it indicates that the body of the text
+or inventory is a forward delta from the referenced revision.  The
+compression base list must have length 0 or 1.
+
+Like packs, indexes are written only once and then unmodified.  A
+GraphIndex builder is a mutable in-memory graph that can be sorted,
+cross-referenced and written out when the write group completes.
+
+There can also be index entries with a value of 'a' for absent.  These
+records exist just to be pointed to in a graph.  This is used, for
+example, to give the revision-parent pointer when the parent revision is
+in a previous pack.
+
+The data content for each record is a knit data chunk.  The knits are
+always unannotated - the annotations must be generated when needed.
+(We'd like to cache/memoize the annotations.)  The data hunks can be
+moved between packs without needing to recompress them.
+
+It is not possible to regenerate an index from the body file, because it
+contains information stored in the knit index that's not in the body.
+(In particular, the per-file graph is only stored in the index.) 
+We would like to change this in a future format.
+
+The lock is a regular LockDir lock.  The lock is only held for a much
+reduced scope, while updating the pack-names file.  The bulk of the
+insertion can be done without the repository locked.  This is an
+implementation detail; the repository user should still call
+``repository.lock_write`` at the regular time but be aware this does not
+correspond to a physical mutex. 
+
+Read locks control caching but do not affect writers.
+
+The newly-added repository write group concept is very important to
+KnitPack repositories.  When ``start_write_group`` is called, a new
+temporary pack is created and all modifications to the repository will 
+go into it until either ``commit_write_group`` or ``abort_write_group``
+is called, at which time it is either finished and moved into place or
+discarded respectively.  Write groups cannot be nested, only one can be
+underway at a time on a Repository instance and they must occur within a
+write lock.
+
+Normally the data for each revision will be entirely within a single
+pack but this is not required.
+
+When a pack is finished, it gets a final name based on the md5 of all
+the data written into the pack body file.
+
+The ``pack-names`` file gives the list of all finished non-obsolete
+packs.  (This should always be the same as the list of files in the
+``packs/`` directory, but the file is needed for readonly http clients
+that can't easily list directories, and it includes other information.)
+The constraint on the ``pack-names`` list is that every file mentioned
+must exist in the ``packs/`` directory.  
+
+In rare cases, when a writer is interrupted, about-to-be-removed packs
+may still be present in the directory but removed from the list.
+
+As well as the list of names, the pack-names file also contains the
+size, in bytes, of each of the four indices.  This is used to bootstrap
+bisection search within the indices.
+
+In normal use, one pack will be created for each commit to a repository.
+This would build up to an inefficient number of files over time, so a
+``repack`` operation is available to recombine them, by producing larger
+files containing data on multiple revisions.  This can be done manually
+by running ``bzr pack``, and it also may happen automatically when a
+write group is committed.
+
+The repacking strategy used at the moment tries to balance not doing too
+much work during commit with not having too many small files left in the
+repository.  The algorithm is roughly this: the total number of
+revisions in the repository is expressed as a decimal number, e.g.
+"532".  Then we'll repack until we have five packs containing a hundred
+revisions each, three packs containing ten revisions each, and two packs
+with single revisions.  This means that each revision will normally
+initially be created in a single-revision pack, then moved to a
+ten-revision pack, then to a 100-pack, and so on.
+
+As with other repositories, in normal use data is only inserted.
+However, in some circumstances we may want to garbage-collect or prune
+existing data, or reconcile indexes.
+
+  vim: tw=72 ft=rest expandtab

=== modified file 'doc/developers/repository.txt'
--- doc/developers/repository.txt	2007-08-01 00:20:37 +0000
+++ doc/developers/repository.txt	2007-10-24 01:52:12 +0000
@@ -271,7 +271,9 @@
 interesting to have it be deterministic based on content, but there are no
 specific problems we have solved by doing that, and doing so would require
 hashing the full file. OTOH hashing the full file is a cheap way to detect
-bit-errors in transfer (such as windows corruption).
+bit-errors in transfer (such as windows corruption). Non-reused file names
+are required for data integrity, as clients having read an index will
+readv at arbitrary times later. 
 
 Discovery of files
 ~~~~~~~~~~~~~~~~~~
@@ -298,7 +300,6 @@
 Choosing compression/delta support
 ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 
-
 Caching and writeing of data
 ============================
 

